#ifndef _GOSTC_ASYM_ELEMS_H_
#define _GOSTC_ASYM_ELEMS_H_

#include "gostc_common.h"

/*********Вспомогательная структура, используемая для сохранения промежуточных значений при целочислном делении******************/
typedef struct _gostc_BUF_ECP
{
	uint32_t sd;
	uint32_t mv;
	uint32_t mu;
	uint32_t mq;
	uint32_t end_v;
	uint32_t end_u;
	uint32_t end_q;
	/*uint32_t beg_u;*/
	uint32_t remainder;
	/*uint32_t szM;*/
	/*uint32_t szN;*/
} gostc_BUF_ECP, *gostc_pBUF_ECP;

/*Точка в проективных координатах эллиптической кривой с привязкой к основанию p.*/

typedef struct _gostc_POINT_ECP
{
	gostc_BUF_ECP proms;
	uint32_t xp[ASYM_DEF_SIZE>>2];
	uint32_t yp[ASYM_DEF_SIZE>>2];
	uint32_t BUF[(ASYM_DEF_SIZE>>1)];
	uint32_t p[ASYM_DEF_SIZE>>2];
	uint32_t Zp[ASYM_DEF_SIZE>>2];
} gostc_POINT_ECP, *gostc_pPOINT_ECP;
/*Точка в проективных координатах эллиптической кривой с привязкой к основанию p. (512 бит)*/
typedef struct _gostc_POINT_ECP64
{
	gostc_BUF_ECP proms;
	uint32_t xp[ASYM_DEF_512_SIZE>>2];
	uint32_t yp[ASYM_DEF_512_SIZE>>2];
	uint32_t BUF[ASYM_DEF_512_SIZE>>1];
	uint32_t p[ASYM_DEF_512_SIZE>>2];
	uint32_t Zp[ASYM_DEF_512_SIZE>>2];
}gostc_POINT_ECP64, gostc_pPOINT_ECP64;

/*Проективные координаты точки эллиптической кривой*/
typedef struct _gostc_ONLY_POINT
{
	uint32_t X[ASYM_DEF_SIZE>>2];
	uint32_t Y[ASYM_DEF_SIZE>>2];
	uint32_t Z[ASYM_DEF_SIZE>>2];
} gostc_ONLY_POINT, *gostc_pONLY_POINT;
/*Проективные координаты точки эллиптической кривой 512 бит.*/
typedef struct _gostc_POINT64
{
	uint32_t X[ASYM_DEF_512_SIZE>>2];
	uint32_t Y[ASYM_DEF_512_SIZE>>2];
	uint32_t Z[ASYM_DEF_512_SIZE>>2];
} gostc_POINT64, *gostc_pPOINT64;


/*Основная структура контекста ГОСТ р 2001: P,Q,A,B,X0,Y0 - криптопараметры. 
P0 - точка с координатами (X0,Y0) 
PM0 - точка с координатами (XМ0,YМ0) в представлении чисел Монтгомери по модулю p
P2M0 - Удвоенная точка PM0 в представлении Монтгомери по модулю p
P3M0 - Утроенная точка PM0 в представлении Монтгомери по модулю p
pK - точки k*P0, где K - особое представление коэффициента кратности чисел k от 0 до 255, выполненное следующим образом:
k = 0, K = 0
k = 1, K = 1
k = 2, K = (1 00000000) hex
k = 3, K = (1 00000001) hex
.....
k = 255, K = (1 00000001 00000001 00000001 00000001 00000001 00000001 00000001) hex
halfP - число (P/2) mod p
M1 - число 1 в представлении Монтгомери по модулю P
R - результат деления 2 в 512 степени! на P (нужно для получения представления Монтгомери)
AM - число А в представлении Монтгомери по модулю p
EpsP - число-шаг представления Монтгомери
*/
typedef struct _gostc_CR_SYSTEM
{
	uint32_t P[ASYM_DEF_SIZE>>2];
	uint32_t Q[ASYM_DEF_SIZE>>2];
	uint32_t A[ASYM_DEF_SIZE>>2];
	uint32_t B[ASYM_DEF_SIZE>>2];
	uint32_t X0[ASYM_DEF_SIZE>>2];
	uint32_t Y0[ASYM_DEF_SIZE>>2];
	gostc_ONLY_POINT P0;
	gostc_ONLY_POINT PM0;
	gostc_ONLY_POINT P2M0;
	gostc_ONLY_POINT P3M0;
	gostc_ONLY_POINT Pk[ASYM_COUNTED_KPOINTS];
	uint32_t halfP[ASYM_DEF_SIZE>>2];
	uint32_t M1[ASYM_DEF_SIZE>>2];
	uint32_t EpsP;
	uint32_t R[ASYM_DEF_SIZE>>2];
	uint32_t AM[ASYM_DEF_SIZE>>2];
} gostc_CR_SYSTEM, *gostc_pCR_SYSTEM;
/*Основная структура 512-битного контекста ГОСТ р 2012: P,Q,A,B,X0,Y0 - криптопараметры. 
P0 - точка с координатами (X0,Y0) 
PM0 - точка с координатами (XМ0,YМ0) в представлении чисел Монтгомери по модулю p
halfP - число (P/2) mod p
M1 - число 1 в представлении Монтгомери по модулю P
R - результат деления 2 в 1024 степени! на P (нужно для получения представления Монтгомери)
AM - число А в представлении Монтгомери по модулю p
EpsP - число-шаг представления Монтгомери
*/
typedef struct __gostc_CR_SYSTEM64
{
	uint32_t P[ASYM_DEF_512_SIZE>>2];
	uint32_t Q[ASYM_DEF_512_SIZE>>2];
	uint32_t A[ASYM_DEF_512_SIZE>>2];
	uint32_t B[ASYM_DEF_512_SIZE>>2];
	uint32_t X0[ASYM_DEF_512_SIZE>>2];
	uint32_t Y0[ASYM_DEF_512_SIZE>>2];
	gostc_POINT64 P0;
	gostc_POINT64 PM0;
	uint32_t halfP[ASYM_DEF_512_SIZE>>2];
	uint32_t M1[ASYM_DEF_512_SIZE>>2];
	uint32_t M3[ASYM_DEF_512_SIZE>>2];
	uint32_t EpsP;
	uint32_t R[(ASYM_DEF_512_SIZE>>2)];
	uint32_t AM[ASYM_DEF_512_SIZE>>2];
}gostc_CR_SYSTEM64, *gostc_pCR_SYSTEM64;
			
/* c = a + b */
/*#########################################################################*/
/*# Здесь и далее разрядность чисел определяется в 32-битных словах       #*/
/*# Возврат функций, если не указано иное: 0-OK, <0-ERROR		  #*/
/*#########################################################################*/

/*  Вычитание 8-разрядных чисел : c = a - b*/
int gostc_Sub_C(uint32_t *a, const uint32_t *b, uint32_t *c);

/*Получение EpsilonQ - шага в представлении числа Монтгомери	*/
/*Если возврат 0, то ошибка					*/
uint32_t gostc_EpsilonQ(uint32_t Q0);

/*Целочисленное деление: q = u/v, u = u%v, k1- разрядность делимого. 
  Разрядность делителя: 8 слов. В параметре s1 сохраняются некоторые промежуточные 
  вычисления во время операции. Не должен быть NULL
  (!)Функция возвращает 0 в случае ошибки; 1 в случае успешного завершения операции; 
  2 - если делитель оказался больше делимого, в этом случае содержимое делимого 
  является остатком от деления, операция считается успешно выполненной, 
  м.б. использовано как warning */
int gostc_Divide_c(uint32_t * u,uint32_t *v, uint32_t * q,uint32_t k1);

/*Сравнение чисел a и b, разрядность: l 32-битных слов*/
/*Возврат: 
 * B > A -  0
 * B < A -  1
 * B == A - 2
 * -1 - ERROR*/
int gostc_Comp_C(const uint32_t* A, const uint32_t* B, const uint32_t l);

/***************************************************************************/
/*Умножение двух чисел: 8-разрядное b и l-разрядное a. Результат пишем в c, разрядность которого 8+l*/
int gostc_Mul(const uint32_t *a, const uint32_t *b,  uint32_t *c,uint32_t l);

/*************************************************************************************/
/*Умножение 8-разрядного числа a на 2 c приведением по модулю p. Результат пишем в n.*/
int gostc_Mul_2C(uint32_t *a,const uint32_t *p,uint32_t *n);

/*************************************************************************************/
/*Умножение 8-разрядного числа a на 3 c приведением по модулю p. Результат пишем в n.*/
int gostc_Mul_3_Mod(uint32_t *a,const uint32_t *p,uint32_t *n);

/************************************************************************************	*/
/*Деление 8-разрядного числа a на 2 c приведением по модулю p. Результат пишем в rez.	*/
/*p2 - число p/2 в представлении Монтгомери						*/
int gostc_div2_C(uint32_t *a,const uint32_t *p2,uint32_t *rez);

/*************************************************************************************/
/*Приведение (c+8)-разрядного числа a по 8-разрядному модулю p в арифметике Монтгомери. 
  e - значение EpsilonQ, полученное с помощью gostc_EpsilonQ.
  Буфер числа a должен содержать один дополнительный разряд на случай переноса, 
  т.о. размер буфера a д.б. (с+8)+1.
  Результатом будет число, помещенное в буфер a начиная с 8-го разряда, 
  при этом разряды а с 0 по 7 будут равны 0 (в случае корректных входных параметров).
  a на входе: xx...xx (16), а после преобразования 0..0y..y
*/
int gostc_ModP_C(uint32_t *a,const uint32_t* p, const uint32_t e, uint32_t c);

/*###################################################################################*/
/*# Операции с 16-разрядными числами   (1 разряд - двойное слово 32 бит.)           #*/
/*###################################################################################*/

/****************************************************************************************/
/*Вычитание: c = a - b*******************************************************************/
int gostc_Sub_C64(uint32_t *a, const uint32_t *b, uint32_t *c);

/*************************************************************************************/
/*Целочисленное деление: q = u/v, u = u%v, k1- разрядность делимого. Разрядность делителя: 16
   (!)Функция возвращает 0 в случае ошибки; 1 в случае успешного завершения операции; 
  2 - если делитель оказался больше делимого, в этом случае содержимое делимого является 
  остатком от деления, операция считается успешно выполненной, м.б. использовано как warning*/
int gostc_Divide_c16(uint32_t *u, uint32_t *v, uint32_t* q,uint32_t k1);

/******************************************************************************/
/*Умножение двух чисел: 16-разрядное b и l-разрядное a. Результат пишем в c, разрядность которого 16+l */
int gostc_Mul16(const uint32_t *a, const uint32_t *b,uint32_t *c,uint32_t l);

/**************************************************************************************/
/*Умножение 16-разрядного числа a на 2 c приведением по модулю p. Результат пишем в n.*/
int gostc_Mul16_2C(uint32_t *a,const uint32_t *p,uint32_t *n);

/************************************************************************************	*/
/*Деление 16-разрядного числа a на 2 c приведением по модулю p. Результат пишем в rez.	*/
/*p2 - число p/2 в представлении Монтгомери						*/
int gostc_div2_64(uint32_t *a,const uint32_t *p2,uint32_t *rez);

/*************************************************************************************/
/*Приведение (c+16)-разрядного числа a по 16-разрядному модулю p в арифметике Монтгомери. 
  e - значение EpsilonQ, полученное с помощью gostc_EpsilonQ
  Буфер числа a должен содержать один дополнительный разряд на случай переноса, 
  т.о. размер буфера a д.б. (с+16)+1.
  Результатом правильного приведения по модулю будет число, помещенное в буфер a начиная 
  с 16-го разряда. Разряды а с 0 по 15 будут равны 0 (в случае корректных входных параметров).
  a на входе: xx...xx (32), а после преобразования 0..0y..y
*/
int gostc_ModP16(uint32_t *a,const uint32_t* p, const uint32_t e, uint32_t c);



#endif

