#include "gostc_asym_mask.h"
#include "gostc_mem.h"

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))



#define MASK_OK 1
static uint32_t Q[] =
{0x3ACCF5B3,0xC59CFC19,0x92976154,0x50FE8A18,0x00000001,0x00000000,0x00000000,0x80000000,
/* Q - test GOST */

 0xb761b893,0x45841b09,0x995ad100,0x6c611070,0xffffffff,0xffffffff,0xffffffff,0xffffffff, 
/* Q - CSPa (CSPSet) */

 0xcc8a198f,0xe497161b,0xf1a624e5,0x5f700cff,0x00000001,0x00000000,0x00000000,0x80000000, 
/* Q - CSPb (Oscar) */

 0x98980bb9,0xf02f3a65,0x1eddfb74,0x582ca351,0x6b41c8aa,0xab1ec85e,0x5a858107,0x9b9f605f};
/* Q - CSPc (CSP2Set) */
						  
#define TF(x) ( x ? 1 : 0 )



/* Masking. n = 
0 - TST Gost
1 - CSPSet (cspa)
2 - OSCAR (cspb)
3 - CSP2Set (cspc)
4 - 512bit-TestParamSet
5 - 512bit-ParamSetA
6 - 512bit-ParamSetB
> 6 - ERROR
*/
int gostc_asym_mask_key(void *key, void* mask,uint32_t n)
{
	uint32_t k[(ASYM_DEF_SIZE>>2)+1];
	uint32_t m[(ASYM_DEF_SIZE>>2)+1];
	uint32_t q[ASYM_DEF_SIZE>>2];
	uint32_t zero[ASYM_DEF_SIZE>>2];
	uint32_t buf[(ASYM_DEF_SIZE>>1)+1];
	uint32_t buf2[(ASYM_DEF_SIZE>>1)+1];
	int compare;
	int rc = !MASK_OK;


	if((key == NULL)||(mask == NULL)||(Q == NULL))
		return !MASK_OK;
	if(n > MASK_TC26B) /*0-3 : 32-bit, 4-6 : 64-bit     */
		return !MASK_OK;
	if(n >= MASK_TST512)
	{
		if(gostc_DoMask64(key, mask, n)==0)
			return !MASK_OK;
		else
			return MASK_OK;
	}

	ZZ(zero);

	if (!memcmp(key, zero, ASYM_DEF_SIZE) || !memcmp(mask, zero, ASYM_DEF_SIZE)) 
		return !MASK_OK;

	ZZ(k);
	ZZ(m);
	ZZ(q);
	ZZ(buf);
	ZZ(buf2);
	ZZS(compare);

	memcpy(k,key,ASYM_DEF_SIZE);
	memcpy(m,mask,ASYM_DEF_SIZE);
	memcpy(q,Q+n*(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
	/* mask = mask(mod Q) */
	if(TF(gostc_Divide_c(m,q,buf,ASYM_DEF_SIZE>>2)) != MASK_OK) {
		rc = !MASK_OK;
		goto end;
	}
	compare = gostc_Comp_C(m,zero,ASYM_DEF_SIZE>>2);
	if(compare <0)/*Error*/
	{
		rc = !MASK_OK;
		goto end;
	}
		
	if(compare == 2)
		m[0] = 1;
	/* buf = key * mask; */
	if(gostc_Mul(k,m,buf,ASYM_DEF_SIZE>>2)<0)
	{
		rc = !MASK_OK;
		goto end;
	}
	/* buf = (buf) mod (Q); */
	if(TF(gostc_Divide_c(buf,q,buf2,ASYM_DEF_SIZE>>1)) != MASK_OK) {
		rc = !MASK_OK;
		goto end;
	}
	memcpy(key,buf,ASYM_DEF_SIZE);
	rc = MASK_OK;
end:
	SZZ(k);
	SZZ(m);
	SZZ(q);
	SZZ(zero);
	SZZ(buf);
	SZZ(buf2);
	SZZS(compare);
	return rc;
}


/* Masking. n = 
0 - TST Gost
1 - CSPSet (cspa)
2 - OSCAR (cspb)
3 - CSP2Set (cspc)
4 - 512bit-TestParamSet
5 - 512bit-ParamSetA
6 - 512bit-ParamSetB
> 6 - ERROR
*/
int gostc_asym_unmask_key(void *key,void*mask,uint32_t n)
{
	uint32_t k[(ASYM_DEF_SIZE>>2)+1];
	uint32_t m[(ASYM_DEF_SIZE>>2)+1];
	uint32_t q[ASYM_DEF_SIZE>>2];
	uint32_t zero[ASYM_DEF_SIZE>>2];
	uint32_t buf[(ASYM_DEF_SIZE>>1)+1];
	uint32_t buf2[(ASYM_DEF_SIZE>>1)+1];
	int compare;
	uint32_t gamma[ASYM_DEF_SIZE>>2];
	/*uint32_t gamma2[ASYM_DEF_SIZE>>2];*/
	int rc = !MASK_OK;


	if((key == NULL)||(mask == NULL)||(Q == NULL))
		return !MASK_OK;
	if(n > MASK_TC26B)
		return !MASK_OK;
	if(n >= MASK_TST512)
	{
		if(gostc_DoUnmask64(key,mask,n)==0)
			return !MASK_OK;
		else
			return MASK_OK;
	}

	ZZ(k);
	ZZ(m);
	ZZ(q);
	ZZ(zero);
	ZZ(buf);
	ZZ(buf2);
	ZZS(compare);
	ZZ(gamma);
	/*ZZ(gamma2);*/

	if (!memcmp(key, zero, ASYM_DEF_SIZE) || !memcmp(mask, zero, ASYM_DEF_SIZE)) 
		return !MASK_OK;
	memcpy(k,key,ASYM_DEF_SIZE);
	memcpy(m,mask,ASYM_DEF_SIZE);
	memcpy(q,Q+n*(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
	
	if (TF(gostc_Divide_c(m,q,buf,ASYM_DEF_SIZE>>2)) != MASK_OK) {
		rc = !MASK_OK;
		goto end;
	}
	compare = gostc_Comp_C(m,zero,ASYM_DEF_SIZE>>2);
	if(compare < 0)/*Error!*/
	{
		rc = !MASK_OK;
		goto end;
	}
	if(compare == 2)
		m[0] = 1;
	if(gostc_BinaryEuclid_C(m,q,gamma)<0)
	{
		rc = !MASK_OK;
		goto end;
	}
	/*if(gostc_BinaryEuclid_C(gamma,q,gamma2)<0)
	{
		rc = !MASK_OK;
		goto end;
	}##Ivan: Лишнее*/
	

	/*gostc_Mul_3c(k,gamma,buf,8);*/
	if(gostc_Mul(k,gamma,buf,ASYM_DEF_SIZE>>2)<0)
	{
		rc =!MASK_OK;
		goto end;
	}
	if (TF(gostc_Divide_c(buf,q,buf2,ASYM_DEF_SIZE>>1)) != MASK_OK) {
		rc = !MASK_OK;
		goto end;
	}
	memcpy(key,buf,ASYM_DEF_SIZE);
	rc = MASK_OK;
end:
	SZZ(k);
	SZZ(m);
	SZZ(q);
	SZZ(zero);
	SZZ(buf);
	SZZ(buf2);
	SZZS(compare);
	SZZ(gamma);
	/*SZZ(gamma2);*/
	return rc;
}

#if 0

#include <time.h>

void Random(char* s,int n)
{
	srand(time(0));
	for(int i =0;i<n;i++)
	{
		s[i] = (char)rand();
	}
}

uint32_t test_d[8]=
{0x91ec3b28,0x1d19ce98,0x49397eee,0x1b60961f,
0xd39a72c1,0x10ed359d,0x789bb9be,0x7a929ade};

void _tmain(int argc, _TCHAR* argv[])
{
	uint32_t Q0[8] = {0x3ACCF5B3,0xC59CFC19,0x92976154,0x50FE8A18,0x00000001,0x00000000,0x00000000,0x80000000};
	uint32_t key[8];
	uint32_t mask[8];
	
	memcpy(key,test_d,32);
	Random((char*)mask,32);
	for(int i=0;i<8;i++)
		printf("Key 0x%8.8x, mask 0x%8.8x\n",key[i],mask[i]);
	printf("params - ?\n");
	int t;
	scanf("%d",&t);
	DoMask(key,mask,t);
	printf("Masking...\n");
	for(int i=0;i<8;i++)
		printf("Key 0x%8.8x, mask 0x%8.8x\n",key[i],mask[i]);

	DoUnmask(key,mask,t);

	printf("Unmasking...\n");
	for(int i=0;i<8;i++)
		printf("Key 0x%8.8x, mask 0x%8.8x\n",key[i],mask[i]);
	printf("Finished\n");
}

#endif

