#include "gostc_asym_opers.h"
#include "gostc_mem.h"

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))




/******************************************************************************/
/*вычисление A*B с приведением по модулю P*/
int gostc_FindEQABModP(uint32_t *A,uint32_t *B,	const uint32_t *p,uint32_t *S)
{
	/*bzero(params->pr+4,72);*/
	uint32_t L[(ASYM_DEF_SIZE>>1) +1];
	uint32_t Divisor[(ASYM_DEF_SIZE>>2)+1];
	gostc_POINT_ECP p1;

	int rc = -1;
	if(!A || !B || !p || !S)
		return rc;/*Ошибка передачи параметров*/

	ZZ(L);
	ZZ(Divisor);
	ZZS(p1);

	memcpy(Divisor, p, ASYM_DEF_SIZE);
	if(gostc_Mul(A,B,L,ASYM_DEF_SIZE>>2)<0)
	{
		rc =-2;
		goto end;
	}
	if(gostc_Divide_c(L,Divisor,p1.BUF,ASYM_DEF_SIZE>>1)==0)
	{
				
		rc=-2;
		goto end;
	}
	memcpy(S,L,ASYM_DEF_SIZE);
	rc = 0;/*OK*/
end:
	SZZ(L);
	SZZ(Divisor);
	return rc;
}

/******************************************************************************/

int gostc_Find16ModP(uint32_t *A,uint32_t *B,const uint32_t *p,uint32_t *S)
{
	uint32_t L[(ASYM_DEF_512_SIZE>>1)+1], Divisor[(ASYM_DEF_512_SIZE>>2)+1];
	int rc = -1;
	gostc_POINT_ECP64 p1;
	if(!A || !B || !p || !S)
		return rc;/*Ошибка передачи параметров*/
	ZZ(L);
	ZZ(Divisor);
	ZZS(p1);
	memcpy(Divisor, p, ASYM_DEF_512_SIZE);

	if(gostc_Mul16(A,B,L,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc =-2;
		goto end;
	}

	if(gostc_Divide_c16(L,Divisor,p1.BUF,ASYM_DEF_512_SIZE>>1)==0)
	{
		rc=-2;
		goto end;
	} 
	memcpy(S,L,ASYM_DEF_512_SIZE);
	rc = 0;/*OK*/
end:
	SZZ(L);
	SZZ(Divisor);
	SZZS(p1);
	return rc;
}


/*****************************************************************************/
/*/Умножение с приведением по модулю p в представлении Монтгомери 8-разр.*/
int gostc_FindEQABModPinM(const gostc_CR_SYSTEM *params,const uint32_t *A,const uint32_t *B,
	const uint32_t *p,uint32_t *S)
{
	uint32_t L[(ASYM_DEF_SIZE>>1)+1];
	int rc = -1;
	if(!params || !A || !B || !p || !S)
		return rc;/*Ошибка передачи параметров*/
	ZZ(L);
	if(gostc_Mul(A,B,L,ASYM_DEF_SIZE>>2)<0)
	{
		rc =-2;
		goto end;
	}
	if(gostc_ModP_C(L,p,params->EpsP,ASYM_DEF_SIZE>>2)<0)
	{
		rc =-2;
		goto end;
	}

	memcpy(S,L+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
	rc = 0;
end:
	SZZ(L);
	return rc;
}
/*****************************************************************************/

/*/Умножение с приведением по модулю p в представлении Монтгомери 16-разр.*/
int gostc_Find16ModPinM(const gostc_CR_SYSTEM64 *params,const uint32_t *A,const uint32_t *B, const uint32_t *p,uint32_t *S)
{
	uint32_t L[(ASYM_DEF_512_SIZE>>1)+1];
	int rc =-1;
	if(!params || !A || !B || !p || !S)
		return rc;/*Ошибка передачи параметров*/
	
	ZZ(L);
	if(gostc_Mul16(A,B,L,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc = -2;
		goto end;
	}
	if(gostc_ModP16(L,p,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc = -2;
		goto end;
	}

	memcpy(S,L+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
	rc=0;
end:
	SZZ(L);
	return rc;
}
/*****************************************************************************/
/*сложение A + B с приведением по модулю P ***/
int gostc_AplusBModP_C(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez)
{
	uint64_t p1 = 0;
	uint64_t p2 = 0;
	uint32_t mod[ASYM_DEF_SIZE>>2]={0};

	int i=0, t=0,rc=-1;
	if(!A || !B || !P || !Rez)
		return rc;/*Ошибка передачи параметров*/
	memcpy(mod,P,ASYM_DEF_SIZE);

	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = p1 + (uint64_t)A[i];
		p1 = p1 + (uint64_t)B[i];
		Rez[i] = (uint32_t)p1;
		p1 = p1 >> 32;
	}
	if(p1 == 1)
		goto subP;
	t = gostc_Comp_C(Rez,mod,ASYM_DEF_SIZE>>2);
	if(t<0)
	{
		rc = -2;
		goto end;
	}	
	if(t != 1) /*не нужно вычитать P*/
	{
		rc = 0;/*OK*/
		goto end;
	}
subP:
	p2 = 0;
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)Rez[i];
		p1 = p1 - (uint64_t)mod[i];/* Отнимаем разряд B */
		p1 = p1 - p2; /* Учитываем заём с пред. разряда */
		Rez[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}
	rc =0;/*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZ(mod);
	SZZS(i);
	SZZS(t);
	return rc;
}
/****************************************************************************/
/*сложение A + B с приведением по модулю P (16-разрядное) ***/
int gostc_AplusBModP64(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez)
{
	uint64_t p1=0,p2=0;
	int i=0,t=0,rc=-1;
	uint32_t mod[ASYM_DEF_512_SIZE>>2]={0};

	if(!A || !B || !P || !Rez)
		return rc;/*Ошибка передачи параметров*/
	memcpy(mod,P,ASYM_DEF_512_SIZE);

	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 = p1 + (uint64_t)A[i];
		p1 = p1 + (uint64_t)B[i];
		Rez[i] = (uint32_t)p1;
		p1 = p1 >> 32;
	}
	if(p1 == 1)
		goto subP64;
	t = gostc_Comp_C(Rez,mod,ASYM_DEF_512_SIZE>>2);
	if(t<0)
	{
		rc = -2;/*OK*/
		goto end;
	}
	if(t != 1) 
	{
		rc = 0;/*OK*/
		goto end;
	}
subP64:
	p2 = 0;
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 = (uint64_t)Rez[i];
		p1 = p1 - (uint64_t)mod[i];/*/Отнимаем разряд B*/
		p1 = p1 - p2; /*Учитываем заём с пред. разряда*/
		Rez[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* //Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}
	rc = 0;/*OK*/
end:
	SZZS(p1);
	SZZS(p2);
	SZZS(i);
	SZZS(t);
	SZZ(mod);
	return rc;
}
/********Вычитание: A = A - B по модулю P****************************************/
int gostc_SimpleSub(uint32_t *A,uint32_t *B,const uint32_t *P)
{
	return gostc_AMinusBModP_C((uint32_t*)A,(uint32_t*)B,P,(uint32_t*)A);
}
/*****************************************************************************/

/******Процедуры вычитания A - B с приведением по модулю P (8-разрядное)   *********/
int gostc_AMinusBModP_C(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez)
{
	uint64_t p1 = 0, p2 = 0;
	int i=0,rc = -1;

	if(!A || !B || !P || !Rez) /*Ошибка передачи параметров*/
		return -1;
	/* First count A - B */
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 = (uint64_t)A[i];
		p1 = p1 - (uint64_t)B[i];/* Subtract digit B */
		p1 = p1 - p2; /* Uchityvajem zajom s pred. razrjada */
		Rez[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1) /* Byl novyj zaem - nado ego sohranit' */
			p2 = 1;
		else
			p2 = 0;
	}
	/* A teper' esli zajom byl - to chislo ushlo v minus, nado pribavit' P	*/
	if(p2 == 0) /* zaema ne bylo - vyhod */
	{
		rc = 0;	
		goto end;
	}
	p1 = 0;
	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		p1 += (uint64_t)Rez[i];
		p1 += (uint64_t)P[i];
		Rez[i] = (uint32_t)p1;
		p1 = p1>>32; /* Uchityvaem vozmozhnyj perenos. */
	}
	rc = 0;
end:
	SZZS(p1);
	SZZS(p2);
	SZZS(i);
	return rc;
}
/************************************************************************************/
/******Процедуры вычитания A - B с приведением по модулю P (16-разрядное)   *********/
int gostc_AMinusBModP64(uint32_t *A,uint32_t *B,const uint32_t *P,uint32_t *Rez)
{
	uint64_t p1=0,p2=0;
	int i=0,rc =-1;

	if(!A || !B || !P || !Rez) /*Ошибка передачи параметров*/
		return rc;

	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 = (uint64_t)A[i];
		p1 = p1 - (uint64_t)B[i];/*///Отнимаем разряд B*/
		p1 = p1 - p2; /* ////Учитываем заём с пред. разряда */
		Rez[i] = (uint32_t)p1;
		if(p1 >>32 == ASYM_MINUS_1)  /*//Был новый заем - надо его сохранить */
			p2 = 1;
		else
			p2 = 0;
	}
/*	/////А теперь если заём был - то число ушло в минус, надо прибавить Р */
	if(p2 == 0) /*///заема не было - выход */
	{	
		rc=0;/*OK*/
		goto end;
	}
	p1 = 0;
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		p1 += (uint64_t)Rez[i];
		p1 += (uint64_t)P[i];
		Rez[i] = (uint32_t)p1;
		p1 = p1>>32; /*///Учитываем возможный перенос. */
	}
	rc = 0;
end:
	SZZS(p1);
	SZZS(p2);
	SZZS(i);
	return rc;
}
/******Процедуры вычисления (A - B)/2 с приведением по модулю P  (8 разрядное) *********/
int gostc_HALFAMinusBModPinM(uint32_t *A,uint32_t *B,const uint32_t *P,const uint32_t *halfP,uint32_t *Rez)
{
	int rc = -1;
	if(!A || !B || !P || !halfP ||!Rez) /*Ошибка передачи параметров*/
		return-1;

	if(gostc_AMinusBModP_C(A,B,P,A)<0)
	{
		rc = -2;
		goto end;
	}	
	if(gostc_div2_C(A,halfP,Rez)<0)
	{
		rc = -2;
		goto end;
	}	
	
	
	rc = 0;	/*OK*/
end:
	
	return rc; 
}
/*****************************************************************************/
/******Процедуры вычисления (A - B)/2 с приведением по модулю P  (16 разрядное) *********/
int gostc_HALFAMinusB64(uint32_t *A,uint32_t *B,const uint32_t *P,const uint32_t *halfP,uint32_t *Rez)
{
	int rc = -1;
	if(!A || !B || !P || !halfP ||!Rez) /*Ошибка передачи параметров*/
		return-1;

	if(gostc_AMinusBModP64(A,B,P,A)<0)
	{
		rc = -2;
		goto end;
	}	
	if(gostc_div2_64(A,halfP,Rez)<0)
	{
		rc = -2;
		goto end;
	}
	rc = 0;/*OK*/
end:
	return rc;
}
/*****************************************************************************/
/****Процедура поиска числа 1/X по модулю P (Бинарный алгоритм Евклида)****/

int gostc_BinaryEuclid_C(uint32_t *a, const uint32_t *p, uint32_t *rez)
{
	uint32_t u[ASYM_DEF_SIZE>>2],v[ASYM_DEF_SIZE>>2],x1[(ASYM_DEF_SIZE>>2)+1],x2[(ASYM_DEF_SIZE>>2)+1],/* pr[9], */one[ASYM_DEF_SIZE>>2], md=0;
	uint64_t p1=0, p2=0;
	int i=0, c1=2, c2=2, compare1=-1,rc = -1;

	if((a == NULL)||(p == NULL)||(rez == NULL))
		return rc;


	for(i=0;i<ASYM_DEF_SIZE>>2;i++)
	{
		u[i] = a[i];
		v[i] = p[i];
		x1[i+1] = 0;
		x2[i+1] = 0;
		/* pr[i+1] = 0; */
		one[i] = 0;
	}
	one[0] = 1;
	x1[0] = 1;
	x2[0] = 0;
	/* pr[0] = 0; */
	md = 0;
	c1 = gostc_Comp_C(one,u,ASYM_DEF_SIZE>>2);
	c2 = gostc_Comp_C(one,v,ASYM_DEF_SIZE>>2);

	while((c1 !=2 )&&(c2 !=2)&&(c1 >= 0) && (c2 >= 0))
	{
		/* poka u - chetnoe - delim u popolam i smotrim iks. */
		if((u[0]&1) == 1)
			goto fin1;
rep1:
		/* Divide U/2 */
		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			u[i] = u[i] >> 1;
			md = u[i+1]&1;
			if(md)
				u[i] += ASYM_ELDER;
		}
		u[(ASYM_DEF_SIZE>>2)-1] = u[(ASYM_DEF_SIZE>>2)-1] >> 1;
		/* Proverjaem x1. Esli x1 chetnoe to delim x1 popolam. */
		if((x1[0]&1) == 1)
			goto adding1;

		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			x1[i] = x1[i] >> 1;
			md = x1[i+1]&1;
			if(md)
				x1[i] += ASYM_ELDER;
		}
		x1[(ASYM_DEF_SIZE>>2)-1] = x1[(ASYM_DEF_SIZE>>2)-1] >> 1;
		goto fin2;
adding1:/* x1 = (x1 + p)/2 */
		p1 = 0;
		for(i=0;i<ASYM_DEF_SIZE>>2;i++)
		{
			p1 += (uint64_t)x1[i];
			p1 += (uint64_t)p[i];
			x1[i] = (uint32_t)p1;
			p1 = p1>>32; /* Uchityvaem vozmozhnyj perenos. */
		}
		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			x1[i] = x1[i] >> 1;
			md = x1[i+1]&1;
			if(md)
				x1[i] += ASYM_ELDER;
		}
		x1[(ASYM_DEF_SIZE>>2)-1] = x1[(ASYM_DEF_SIZE>>2)-1] >> 1;
		if(p1)
			x1[7] += ASYM_ELDER;

fin2:
		if((u[0]&1) == 0)
			goto rep1;

fin1:
		/* Check v */
		if((v[0]&1) == 1)
			goto fin3;
rep3:
		/* Divide V/2 */
		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			v[i] = v[i] >> 1;
			md = v[i+1]&1;
			if(md)
				v[i] += ASYM_ELDER;
		}
		v[(ASYM_DEF_SIZE>>2)-1] = v[(ASYM_DEF_SIZE>>2)-1] >> 1;
		/* Check x2. If x2 is even, divide x2/2 */
		if((x2[0]&1) == 1)
			goto adding3;
		
		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			x2[i] = x2[i] >> 1;
			md = x2[i+1]&1;
			if(md)
				x2[i] += ASYM_ELDER;
		}
		x2[(ASYM_DEF_SIZE>>2)-1] = x2[(ASYM_DEF_SIZE>>2)-1] >> 1;
		goto fin4;
adding3:/* x2 = (x2 + p)/2 */
		p2 = 0;
		for(i=0;i<ASYM_DEF_SIZE>>2;i++)
		{
			p2 += (uint64_t)x2[i];
			p2 += (uint64_t)p[i];
			x2[i] = (uint32_t)p2;
			p2 = p2>>32; /* Uchityvaem vozmozhnyj perenos. */
		}
		md = 0;
		for(i=0;i<(ASYM_DEF_SIZE>>2)-1;i++)
		{
			x2[i] = x2[i] >> 1;
			md = x2[i+1]&1;
			if(md)
				x2[i] += ASYM_ELDER;
		}
		x2[(ASYM_DEF_SIZE>>2)-1] = x2[(ASYM_DEF_SIZE>>2)-1] >> 1;
		if(p2)
			x2[(ASYM_DEF_SIZE>>2)-1] += ASYM_ELDER;

fin4:
		if((v[0]&1) == 0)
			goto rep3;

fin3:
		/* Compare U and V */
		compare1 = gostc_Comp_C(u,v,ASYM_DEF_SIZE>>2);
		if(compare1 <0)/*ОШибка*/
		{
			rc = -2;
			goto end;
		}	
		if((compare1 == 1)||(compare1 ==2))
		{
			/* u = u -v; */
			if(gostc_AMinusBModP_C((uint32_t*)u,(uint32_t*)v,(uint32_t*)p,(uint32_t*)u)<0)
				{rc =-3; goto end;}
			/* x1 = x1 - x2 */
			if(gostc_AMinusBModP_C((uint32_t*)x1,(uint32_t*)x2,(uint32_t*)p,(uint32_t*)x1)<0)
				{rc =-3; goto end;}

		}
		else
		{
			/* v = v -u; */
			if(gostc_AMinusBModP_C((uint32_t*)v,(uint32_t*)u,(uint32_t*)p,(uint32_t*)v)<0)
				{rc =-3; goto end;}

			/* x2 = x2 - x1 */
			if(gostc_AMinusBModP_C((uint32_t*)x2,(uint32_t*)x1,(uint32_t*)p,(uint32_t*)x2)<0)
				{rc =-3; goto end;}
		}
		c1 = gostc_Comp_C(one,u,ASYM_DEF_SIZE>>2);
		c2 = gostc_Comp_C(one,v,ASYM_DEF_SIZE>>2);
	}
	if((c1 == 2 && c2 == 2)|| (c1 < 0) || (c2 < 0))/*Вырожденные случаи или ошибка*/
	{
		rc = -2;
		goto end;
	}
	if(c1 == 2)
	{
		memcpy(rez,x1,ASYM_DEF_SIZE);
	}
	else
	{
		memcpy(rez,x2,ASYM_DEF_SIZE);
	}
	rc = 0;/*OK*/
end:
	SZZ(u);
	SZZ(v);
	SZZ(x1);
	SZZ(x2);
	SZZ(one);
	SZZS(md);
	SZZS(p1);
	SZZS(p2);
	SZZS(i);
	SZZS(c1);
	SZZS(c2);
	SZZS(compare1);

	return rc;
}
/*****************************************************************************/

/****Процедура поиска числа 1/X по модулю P (Бинарный алгоритм Евклида), 16-разрядное****/

int gostc_BinaryEuclid64(uint32_t *a, const uint32_t *p, uint32_t *rez)
{
	uint32_t u[ASYM_DEF_512_SIZE>>2],v[ASYM_DEF_512_SIZE>>2],x1[(ASYM_DEF_512_SIZE>>2)+1],x2[(ASYM_DEF_512_SIZE>>2)+1],one[ASYM_DEF_512_SIZE>>2],md=0;
	uint64_t p1=0, p2=0;
	int i=0, c1=2, c2=2, compare1=-1,rc=-1;
	
	if((a == NULL)||(p == NULL)||(rez == NULL))
		return -1;

	ZZ(one);
	for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
	{
		u[i] = a[i];
		v[i] = p[i];
		x1[i+1] = 0;
		x2[i+1] = 0;
		one[i] = 0;
	}
	one[0] = 1;
	x1[0] = 1;
	x2[0] = 0;
	md = 0;
	c1 = gostc_Comp_C(one,u,ASYM_DEF_512_SIZE>>2);
	c2 = gostc_Comp_C(one,v,ASYM_DEF_512_SIZE>>2);
	
	while((c1 !=2 )&&(c2 !=2)&&(c1 >= 0) && (c2 >= 0))
	{
		/*//пока u - четное - делим u пополам и смотрим икс.*/
		if((u[0]&1) == 1)
			goto fin164;
rep164:
		/*///Делим U пополам*/
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			u[i] = u[i] >> 1;
			md = u[i+1]&1;
			if(md)
				u[i] += ASYM_ELDER;
		}
		u[(ASYM_DEF_512_SIZE>>2)-1] = u[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
/*		///Проверяем x1. Если x1 четное то делим x1 пополам.*/
		if((x1[0]&1) == 1)
			goto adding164;
		
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			x1[i] = x1[i] >> 1;
			md = x1[i+1]&1;
			if(md)
				x1[i] += ASYM_ELDER;
		}
		x1[(ASYM_DEF_512_SIZE>>2)-1] = x1[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
		goto fin264;
adding164: /*///x1 = (x1 + p)/2*/
		p1 = 0;
		for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
		{
			p1 += (uint64_t)x1[i];
			p1 += (uint64_t)p[i];
			x1[i] = (uint32_t)p1;
			p1 = p1>>32;/*///Учитываем возможный перенос.*/
		}
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			x1[i] = x1[i] >> 1;
			md = x1[i+1]&1;
			if(md)
				x1[i] += ASYM_ELDER;
		}
		x1[(ASYM_DEF_512_SIZE>>2)-1] = x1[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
		if(p1)
			x1[(ASYM_DEF_512_SIZE>>2)-1] += ASYM_ELDER;

fin264:
		if((u[0]&1) == 0)
			goto rep164;

fin164:
		/*//Аналогично проверяем v*/
		if((v[0]&1) == 1)
			goto fin364;
rep364:
		/*///Делим V пополам*/
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			v[i] = v[i] >> 1;
			md = v[i+1]&1;
			if(md)
				v[i] += ASYM_ELDER;
		}
		v[(ASYM_DEF_512_SIZE>>2)-1] = v[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
		/*///Проверяем x2. Если x2 четное то делим x2 пополам.*/
		if((x2[0]&1) == 1)
			goto adding364;
		
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			x2[i] = x2[i] >> 1;
			md = x2[i+1]&1;
			if(md)
				x2[i] += ASYM_ELDER;
		}
		x2[(ASYM_DEF_512_SIZE>>2)-1] = x2[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
		goto fin464;
adding364:/*///x2 = (x2 + p)/2*/
		p2 = 0;
		for(i=0;i<ASYM_DEF_512_SIZE>>2;i++)
		{
			p2 += (uint64_t)x2[i];
			p2 += (uint64_t)p[i];
			x2[i] = (uint32_t)p2;
			p2 = p2>>32;/*///Учитываем возможный перенос.*/
		}
		md = 0;
		for(i=0;i<(ASYM_DEF_512_SIZE>>2)-1;i++)
		{
			x2[i] = x2[i] >> 1;
			md = x2[i+1]&1;
			if(md)
				x2[i] += ASYM_ELDER;
		}
		x2[(ASYM_DEF_512_SIZE>>2)-1] = x2[(ASYM_DEF_512_SIZE>>2)-1] >> 1;
		if(p2)
			x2[(ASYM_DEF_512_SIZE>>2)-1] += ASYM_ELDER;

fin464:
		if((v[0]&1) == 0)
			goto rep364;

fin364:
		/*//Сравниваем u и v*/
		compare1 = gostc_Comp_C(u,v,ASYM_DEF_512_SIZE>>2);
		if(compare1 <0)/*ОШибка*/
		{
			rc = -2;
			goto end;
		}		
		if((compare1 == 1)||(compare1 ==2))
		{
			/*u = u -v;*/
			if(gostc_AMinusBModP64(u,v,p,u)<0)
			{
				rc = -2;
				goto end;
			}
			/*//x1 = x1 - x2*/
			if(gostc_AMinusBModP64(x1,x2,p,x1)<0)
			{
				rc = -2;
				goto end;
			}
		}
		else
		{
			/*//v = v -u;*/
			if(gostc_AMinusBModP64(v,u,p,v)<0)
			{
				rc = -2;
				goto end;
			}
			/*//x2 = x2 - x1*/
			if(gostc_AMinusBModP64(x2,x1,p,x2)<0)
			{
				rc = -2;
				goto end;
			}
		}
		c1 = gostc_Comp_C(one,u,ASYM_DEF_512_SIZE>>2);
		c2 = gostc_Comp_C(one,v,ASYM_DEF_512_SIZE>>2);

	}
	if((c1 == 2 && c2 == 2)|| (c1 < 0) || (c2 < 0))/*Вырожденные случаи или ошибка*/
	{
		rc = -2;
		goto end;
	}
	if(c1 == 2)
	{
		memcpy(rez,x1,ASYM_DEF_512_SIZE);
	}
	else
	{
		memcpy(rez,x2,ASYM_DEF_512_SIZE);
	}
	rc = 0;/*OK*/
end:
	SZZ(u);
	SZZ(v);
	SZZ(x1);
	SZZ(x2);
	SZZ(one);
	SZZS(md);
	SZZS(p1);
	SZZS(p2);
	SZZS(i);
	SZZS(c1);
	SZZS(c2);
	SZZS(compare1);

	return rc; 
}

/*******************************************************************************/

/*Cложение двух точек в проективных координатах pa и pb, 
результат сохраняется в prez (возврат: 0 - ОК,  <0 - ошибка) */


int gostc_AddingInProjectiveWInM(gostc_pONLY_POINT pa, gostc_pONLY_POINT pb, 
	gostc_pONLY_POINT pRez, const gostc_CR_SYSTEM *params)
{
	int rc = -2;	/*Ошибка вычислений*/
	uint32_t L[ASYM_PROM_COUNT][ASYM_DEF_SIZE>>2];/*Для промежуточных результатов*/
	

	if(!pa || !pb || !pRez || !params)
	{
		rc = -1;
		return rc;/*Ошибка передачи параметров*/
	}
	ZZ(L);
	if(memcmp(pb->Z,params->M1,ASYM_DEF_SIZE)!=0)/*Если Z-координата второй точки равна 1, то процедура урощается*/
	{
		if(gostc_FindEQABModPinM(params,pa->Z,pa->Z,params->P,L[1])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,pb->Z,pb->Z,params->P,L[2])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,pa->X,L[2],params->P,L[3])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,pb->X,L[1],params->P,L[4])<0)
			return rc;
		if(gostc_AMinusBModP_C(L[3],L[4],params->P,L[5])<0)			
			return rc;
		if(gostc_AplusBModP_C(L[3],L[4],params->P,L[6])<0)/*lam6 = lam3+lam4*/
			return rc;
	/*//////////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,pa->Y,pb->Z,params->P,L[7])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,L[7],L[2],params->P,L[3])<0)	
			return rc;
		if(gostc_FindEQABModPinM(params,pb->Y,pa->Z,params->P,L[7])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,L[7],L[1],params->P,L[4])<0)
			return rc;	
		if(gostc_AMinusBModP_C(L[3],L[4],params->P,L[1])<0)/*lam1 = lam3-lam4*/
			return rc;
		if(gostc_AplusBModP_C(L[3],L[4],params->P,L[2])<0)/*lam2 = lam3+lam4*/
			return rc;
	/*//////////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,L[5],L[5],params->P,L[3])<0)	
			return rc;
		if(gostc_FindEQABModPinM(params,L[3],L[6],params->P,L[4])<0)	
			return rc;
	/*//////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,L[1],L[1],params->P,L[7])<0)	
			return rc;
		if(gostc_AMinusBModP_C(L[7],L[4],params->P,pRez->X)<0)/*X3 = lam1*lam1 - lam4*/
			return rc;
	/*//////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,pb->Z,pa->Z,params->P,L[8])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,L[5],L[8],params->P,pRez->Z)<0)
			return rc;	
		if(gostc_Mul_2C(pRez->X,params->P,L[8])<0)
			return rc;
		if(gostc_AMinusBModP_C(L[4],L[8],params->P,L[6])<0)/*lam6 = lam4 - 2*X3*/
			return rc;
	/*///////////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,L[1],L[6],params->P,L[7])<0)	
			return rc;
		if(gostc_FindEQABModPinM(params,L[2],L[3],params->P,L[8])<0)
			return rc;	
		if(gostc_FindEQABModPinM(params,L[8],L[5],params->P,L[9])<0)	
			return rc;
		if(gostc_HALFAMinusBModPinM(L[7],L[9],params->P,params->halfP,pRez->Y)<0)/*lam1*lam6 - lam2*lam3*lam5*/
			return rc;
	}
	else
	{
		if(gostc_FindEQABModPinM(params,pa->Z,pa->Z,params->P,L[1])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,pb->X,L[1],params->P,L[4])<0)
			return rc;
		if(gostc_AMinusBModP_C(pa->X,L[4],params->P,L[5])<0)/*lam5 = lam3-lam4*/
			return rc;
		if(gostc_AplusBModP_C(pa->X,L[4],params->P,L[6])<0)/*lam6 = lam3+lam4*/
			return rc;
	/*//////////////////////////////////////////////////////////////////////////////// */
		/*////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,pb->Y,pa->Z,params->P,L[7])<0)
			return rc;
		if(gostc_FindEQABModPinM(params,L[7],L[1],params->P,L[4])<0)
			return rc;	
		if(gostc_AMinusBModP_C(pa->Y,L[4],params->P,L[1])<0)/*lam1 = lam3-lam4*/
			return rc;
		if(gostc_AplusBModP_C(pa->Y,L[4],params->P,L[2])<0)/*lam2 = lam3+lam4*/
			return rc;
	/*//////////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,L[5],L[5],params->P,L[3])<0)
			return rc;	
		if(gostc_FindEQABModPinM(params,L[3],L[6],params->P,L[4])<0)
			return rc;	
		if(gostc_FindEQABModPinM(params,L[1],L[1],params->P,L[7])<0)
			return rc;	
		if(gostc_AMinusBModP_C(L[7],L[4],params->P,pRez->X)<0)/*X3 = lam1*lam1 - lam4*/
			return rc;
		if(gostc_FindEQABModPinM(params,L[5],pa->Z,params->P,pRez->Z)<0)
			return rc;
/*		Mul_4_Mod(L[5],pa->Z,params->pr,8,params->P,params->EpsP,pRez->Z);*/
	/*/////////////////////////////////////////////////////////////////////// */
		if(gostc_Mul_2C(pRez->X,params->P,L[8])<0)
			return rc;
		if(gostc_AMinusBModP_C(L[4],L[8],params->P,L[6])<0)/*lam6 = lam4 - 2*X3*/
			return rc;
	/*///////////////////////////////////////////////////////////////////////////////// */
		if(gostc_FindEQABModPinM(params,L[1],L[6],params->P,L[7])<0)
			return rc;	
		if(gostc_FindEQABModPinM(params,L[2],L[3],params->P,L[8])<0)
			return rc;	
		if(gostc_FindEQABModPinM(params,L[8],L[5],params->P,L[9])<0)
			return rc;	
		if(gostc_HALFAMinusBModPinM(L[7],L[9],params->P,params->halfP,pRez->Y)<0)/* lam1*lam6 - lam2*lam3*lam5 */
			return rc;
	}
	rc = 0;
	SZZ(L);
	return rc;/*OK*/
}


/*****************************************************************************/
int gostc_AddInProj64M(gostc_pPOINT64 pa,gostc_pPOINT64 pb,gostc_pPOINT64 pRez,const gostc_CR_SYSTEM64* params)
{
	int rc = -2;	
	uint32_t L[ASYM_PROM_COUNT][ASYM_DEF_512_SIZE>>2];
	if(!pa || !pb || !pRez || !params)
	{
		rc = -1;
		return rc;/*Ошибка передачи параметров*/
	}

	ZZ(L);

	if(memcmp(pb->Z,params->M1,ASYM_DEF_512_SIZE)!=0)/*Если Z-координата второй точки равна 1, то процедура урощается*/
	{
		if(gostc_Find16ModPinM(params,pa->Z,pa->Z,params->P,L[1])<0)
			return rc;
		if(gostc_Find16ModPinM(params,pb->Z,pb->Z,params->P,L[2])<0)
			return rc;
		if(gostc_Find16ModPinM(params,pa->X,L[2],params->P,L[3])<0)
			return rc;
		if(gostc_Find16ModPinM(params,pb->X,L[1],params->P,L[4])<0)
			return rc;
		if(gostc_AMinusBModP64(L[3],L[4],params->P,L[5])<0)/*//lam5 = lam3-lam4*/
			return rc;
		if(gostc_AplusBModP64(L[3],L[4],params->P,L[6])<0)/*//lam6 = lam3+lam4*/
			return rc;
/*	////////////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,pa->Y,pb->Z,params->P,L[7])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[7],L[2],params->P,L[3])<0)
			return rc;	
		/*///////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,pb->Y,pa->Z,params->P,L[7])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[7],L[1],params->P,L[4])<0)
			return rc;	
/*	///////////////////////////////////////////////////////////////////////////// */
		if(gostc_AMinusBModP64(L[3],L[4],params->P,L[1])<0)
			return rc;
		if(gostc_AplusBModP64(L[3],L[4],params->P,L[2])<0)
			return rc;
/*	///////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,L[5],L[5],params->P,L[3])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[3],L[6],params->P,L[4])<0)
			return rc;	
	/*///////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,L[1],L[1],params->P,L[7])<0)
			return rc;	
		if(gostc_AMinusBModP64(L[7],L[4],params->P,pRez->X)<0)
			return rc;
	/*///////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,pb->Z,pa->Z,params->P,L[8])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[5],L[8],params->P,pRez->Z)<0)
			return rc;	
/*	//////////////////////////////////////////////////////////////////////// */
		if(gostc_Mul16_2C(pRez->X,params->P,L[8])<0)
			return rc;
		if(gostc_AMinusBModP64(L[4],L[8],params->P,L[6])<0)
			return rc;
/*////////////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,L[1],L[6],params->P,L[7])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[2],L[3],params->P,L[8])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[8],L[5],params->P,L[9])<0)
			return rc;	
		if(gostc_HALFAMinusB64(L[7],L[9],params->P,params->halfP,pRez->Y)<0)
			return rc;
	}
	else
	{
		if(gostc_Find16ModPinM(params,pa->Z,pa->Z,params->P,L[1])<0)
			return rc;
		if(gostc_Find16ModPinM(params,pb->X,L[1],params->P,L[4])<0)
			return rc;
/*////////////////////////////////////////////////////////////////////// */
		if(gostc_AMinusBModP64(pa->X,L[4],(uint32_t*)params->P,(uint32_t*)L[5])<0)
			return rc;
		if(gostc_AplusBModP64(pa->X,L[4],params->P,L[6])<0)
			return rc;
/*///////////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,pb->Y,pa->Z,params->P,L[7])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[7],L[1],params->P,L[4])<0)
			return rc;	
		if(gostc_AMinusBModP64(pa->Y,L[4],params->P,L[1])<0)
			return rc;
		if(gostc_AplusBModP64(pa->Y,L[4],params->P,L[2])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[5],L[5],params->P,L[3])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[3],L[6],params->P,L[4])<0)
			return rc;
		if(gostc_Find16ModPinM(params,L[1],L[1],params->P,L[7])<0)
			return rc;	
		if(gostc_AMinusBModP64(L[7],L[4],params->P,pRez->X)<0)
			return rc;
/*///////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,L[5],pa->Z,params->P,pRez->Z)<0)
			return rc;
		if(gostc_Mul16_2C(pRez->X,params->P,L[8])<0)
			return rc;
		if(gostc_AMinusBModP64(L[4],L[8],params->P,L[6])<0)
			return rc;
/*////////////////////////////////////////////////////////////////////////////////// */
		if(gostc_Find16ModPinM(params,L[1],L[6],params->P,L[7])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[2],L[3],params->P,L[8])<0)
			return rc;	
		if(gostc_Find16ModPinM(params,L[8],L[5],params->P,L[9])<0)
			return rc;	
		if(gostc_HALFAMinusB64(L[7],L[9],params->P,params->halfP,pRez->Y)<0)
			return rc;
	}
	rc = 0;
	SZZ(L);
	return rc;/*OK*/
}



/*****************************************************************************/
/*Удвоение точки pa в проективных координатах: 
Если pRez == NULL, то результат сохраняется в pa
возврат: 0 - ОК, <0 - ошибка*/

int gostc_DoubleInProjectiveWM(const gostc_CR_SYSTEM *params, gostc_pONLY_POINT pa, 
	gostc_pONLY_POINT pRez)
{
	gostc_ONLY_POINT prom;
	uint32_t L[ASYM_PROM_COUNT][ASYM_DEF_SIZE>>2];/*Для промежуточных результатов*/
	int rc = -2;
	if(!params || !pa)
	{
		rc = -1;
		return rc;/*Ошибка передачи параметров!*/
	}
	ZZS(prom);
	ZZ(L);

	if(gostc_Mul_2C(pa->Y,params->P,L[1])<0)
		return rc;
	if(gostc_FindEQABModPinM(params,L[1],L[1],params->P,L[2])<0)
		return rc;	
	if(gostc_FindEQABModPinM(params,pa->X,L[2],params->P,L[3])<0)
		return rc;	
	if(gostc_FindEQABModPinM(params,pa->X,pa->X,params->P,L[5])<0)
		return rc;	
	if(gostc_Mul_3_Mod(L[5],params->P,L[6])<0)
		return rc;
	if(gostc_FindEQABModPinM(params,pa->Z,pa->Z,params->P,L[7])<0)
		return rc;
	if(gostc_FindEQABModPinM(params,L[7],L[7],params->P,L[7])<0)
		return rc;	
	if(gostc_FindEQABModPinM(params,params->AM,L[7],params->P,L[8])<0)
		return rc;
	if(gostc_AplusBModP_C(L[6],L[8],params->P,L[4])<0)/* lam4 = aZ*Z*Z*Z + 3X*X */
		return rc;
	/*////////////////////////////////////////////////////////// */
	if(gostc_FindEQABModPinM(params,L[2],L[2],params->P,L[5])<0)
		return rc;	
	if(gostc_div2_C(L[5],params->halfP,L[2])<0)
		return rc;
	/*////////////////////////////////////////////////////////////////////// */
	if(gostc_FindEQABModPinM(params,pa->Z,L[1],params->P,prom.Z)<0)
		return rc;	
	/*/////////////////////////////////////////////////////////////////// */
	if(gostc_FindEQABModPinM(params,L[4],L[4],params->P,prom.X)<0)
		return rc;	
	if(gostc_Mul_2C(L[3],params->P,L[6])<0)
		return rc;
	if(gostc_SimpleSub(prom.X,L[6],params->P)<0)/*X3 = lam4*lam4 - 2*lam3*/
		return rc;
	if(gostc_AMinusBModP_C(L[3],prom.X,params->P,L[10])<0)/*lam3 - X3*/
		return rc;
	if(gostc_FindEQABModPinM(params,L[4],L[10],params->P,prom.Y)<0)
		return rc;	
	if(gostc_SimpleSub(prom.Y,L[2],params->P)<0)
		return rc;

	if(pRez == NULL)
		memcpy(pa,&prom,ASYM_POINTSZ_256BIT*sizeof(uint32_t));
	else
		memcpy(pRez,&prom,ASYM_POINTSZ_256BIT*sizeof(uint32_t));
	rc =0;
	SZZS(prom);
	SZZ(L);
	return rc;/*OK*/
}	

/*****************************************************************************/

/*//Удвоение точки эллиптической кривой 64-байтной разрядности*/

int gostc_DoubleInProj64M(const gostc_CR_SYSTEM64* params, gostc_pPOINT64 pa, gostc_pPOINT64 pRez)
{
	int rc = -2;
	gostc_POINT64 prom;
	uint32_t L[ASYM_PROM_COUNT][ASYM_DEF_512_SIZE>>2];	
	if(!params || !pa)
	{
		rc = -1;
		return rc;/*Ошибка передачи параметров!*/
	}
	ZZS(prom);
	ZZ(L);
	if(gostc_Mul16_2C(pa->Y,params->P,L[1])<0)
		return rc;
	if(gostc_Find16ModPinM(params,L[1],L[1],params->P,L[2])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,pa->X,L[2],params->P,L[3])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,pa->X,pa->X,params->P,L[5])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,L[5],params->M3,params->P,L[6])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,pa->Z,pa->Z,params->P,L[7])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,L[7],L[7],params->P,L[7])<0)
		return rc;	
	if(gostc_Find16ModPinM(params,params->AM,L[7],params->P,L[8])<0)
		return rc;	
	if(gostc_AplusBModP64(L[6],L[8],params->P,L[4])<0)/*/lam4 = aZ*Z*Z*Z + 3X*X*/
		return rc;
	/*///////////////////////////////////////////////////////////*/
	if(gostc_Find16ModPinM(params,L[2],L[2],params->P,L[5])<0)
		return rc;	
	if(gostc_div2_64(L[5],params->halfP,L[2])<0)
		return rc;
	if(gostc_Find16ModPinM(params,pa->Z,L[1],params->P,prom.Z)<0)
		return rc;	
	if(gostc_Find16ModPinM(params,L[4],L[4],params->P,prom.X)<0)
		return rc;	
	if(gostc_Mul16_2C(L[3],params->P,L[6])<0)
		return rc;
	if(gostc_AMinusBModP64(prom.X,L[6],params->P,prom.X)<0)/*//X3 = lam4*lam4 - 2*lam3*/
		return rc;
	if(gostc_AMinusBModP64(L[3],prom.X,params->P,L[10])<0)/*//lam3 - X3*/
		return rc;
	if(gostc_Find16ModPinM(params,L[4],L[10],params->P,prom.Y)<0)
		return rc;	
	if(gostc_AMinusBModP64(prom.Y,L[2],params->P,prom.Y)<0)
		return rc;
	
	if(pRez == NULL)
		memcpy(pa,&prom,ASYM_POINTSZ_512BIT*sizeof(uint32_t));
	else
		memcpy(pRez,&prom,ASYM_POINTSZ_512BIT*sizeof(uint32_t));

	rc=0;
	SZZS(prom);
	SZZ(L);
	return rc;/*0 - OK*/	
}	


/************************************************************************/
/*Вычисление кратной точки! kP, кде k - представляется числом mas       */
/*Возврат: 0 - ОК, <0 - ошибка вычисления				*/
int gostc_FindKrPointInM(const gostc_CR_SYSTEM *params,uint8_t *mas,int noconvert, gostc_POINT_ECP *p1)
{
	uint32_t k[ASYM_BIT_SZ];
	uint32_t pmas[ASYM_DEF_SIZE>>2];
	int32_t j=0, i=0;
	uint32_t kn[ASYM_DEF_SIZE>>2];
	uint8_t kr[ASYM_DEF_SIZE];
	uint32_t n = ASYM_DEF_SIZE-1;
	gostc_ONLY_POINT P;
	int rc = -1;
	int32_t l = 0;
	uint32_t NZ[(ASYM_DEF_SIZE>>1)+1];
	uint32_t gamma[ASYM_DEF_SIZE>>2];
	
	if(!params || !mas || !p1)
	{
		rc = -1;/*Ошибка передачи параметров!*/
		return rc;
	}

	ZZ(k);
	ZZ(kn);
	ZZ(kr);
	ZZS(P);

	memcpy(pmas,mas,ASYM_DEF_SIZE);
	
	for(j=0; j<ASYM_DEF_SIZE>>2; j++)
	{
		l = 0;
		while(pmas[j]>1)
		{
			k[j*32+l] = pmas[j]&1;
			pmas[j] = pmas[j]/2;
			l++;
		}
		k[j*32+l] = pmas[j];	
	}	
	for(j=0;j<ASYM_DEF_SIZE;j++)
	{
		kn[0] = k[j];
		kn[1] = k[j+ASYM_DEF_SIZE];
		kn[2] = k[j+ASYM_DEF_SIZE*2];	
		kn[3] = k[j+ASYM_DEF_SIZE*3];
		kn[4] = k[j+ASYM_DEF_SIZE*4];
		kn[5] = k[j+ASYM_DEF_SIZE*5];
		kn[6] = k[j+ASYM_DEF_SIZE*6];
		kn[7] = k[j+ASYM_DEF_SIZE*7];
		kr[j] = kn[0] + 2*kn[1]+ 4*kn[2]+8*kn[3]+16*kn[4]
		+ 32*kn[5]+64*kn[6]+128*kn[7];
	}
	while(n > 0 && kr[n]==0)
		n--;
	if(n==0) {
		rc = -1;/*Вырожденный случай: нулевая кратная точка. Это ошибка в данном случае*/
		goto end;
	}
	
	i = n-1;
	
	/*step-1*/
	memcpy(&P,&params->Pk[kr[n]],ASYM_POINTSZ_256BIT*sizeof(uint32_t));
	
	/*step-2 */
	while(i >= 0)
	{
		if(gostc_DoubleInProjectiveWM(params,&P,NULL)<0)
		{
			rc = -2;
			goto end;
		}
		if(kr[i]!=0)
		{
	/*сложение*/
			if(gostc_AddingInProjectiveWInM(&P,(gostc_pONLY_POINT)&params->Pk[kr[i]],&P,params)<0)
			{
				rc = -2;
				goto end;
			}				
		}
		i--;
	}
	/*/Нахождение обратных чисел... (Step - 9)*/
	if(!noconvert)
	{
		ZZ(NZ);
		ZZ(gamma);
		memcpy(NZ,P.Z,ASYM_DEF_SIZE);
		if(gostc_ModP_C(NZ,params->P,params->EpsP,ASYM_DEF_SIZE>>2)<0)
		{
			rc =-3;
			goto end;
		}
		memcpy(P.Z,NZ+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
		ZZ(NZ);

		memcpy(NZ,P.X,ASYM_DEF_SIZE);
		if(gostc_ModP_C(NZ,params->P,params->EpsP,ASYM_DEF_SIZE>>2)<0)
		{
			rc =-3;
			goto end;
		}
		memcpy(P.X,NZ+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
		ZZ(NZ);

		memcpy(NZ,P.Y,ASYM_DEF_SIZE);
		if(gostc_ModP_C(NZ,params->P,params->EpsP,ASYM_DEF_SIZE>>2)<0)
		{
			rc =-3;
			goto end;
		}
		memcpy(P.Y,NZ+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
		
		if(gostc_BinaryEuclid_C(P.Z,params->P,gamma)<0)
		{
			rc = -4;
			goto end;
		}
		/*///////////step 9////////////////////////// */
		if(gostc_FindEQABModP(gamma,P.X,params->P,P.X)<0)
			{rc =-5; goto end;}
		if(gostc_FindEQABModP(gamma,P.X,params->P,p1->xp)<0)
			{rc =-5; goto end;}
		if(gostc_FindEQABModP(gamma,P.Y,params->P,P.Y)<0)
			{rc =-5; goto end;}
		if(gostc_FindEQABModP(gamma,P.Y,params->P,P.Y)<0)
			{rc =-5; goto end;}
		if(gostc_FindEQABModP(gamma,P.Y,params->P,p1->yp)<0)
			{rc =-5; goto end;}
		rc = 0;
		goto end;
	}
	else
	{
		memcpy(p1->xp,P.X,ASYM_DEF_SIZE);
		memcpy(p1->yp,P.Y,ASYM_DEF_SIZE);
		memcpy(p1->Zp,P.Z,ASYM_DEF_SIZE);
				
		rc = 0;
	}
end:
	SZZ(k);
	SZZ(pmas);
	SZZS(i);
	SZZS(j);
	SZZ(kn);
	SZZ(kr);
	SZZS(n);
	SZZS(P);
	SZZS(l);
	SZZ(NZ);
	SZZ(gamma);
	return rc;
}

/*****************************************************************************/
int gostc_FindKrPoint64(const gostc_CR_SYSTEM64* params,uint8_t* mas,int noconvert, gostc_POINT_ECP64 *p1)
{
	uint32_t k[ASYM_BIT_512_SZ+2],gamma[ASYM_DEF_512_SIZE>>2]; 
	uint32_t pmas[ASYM_DEF_512_SIZE>>2],NZ[(ASYM_DEF_512_SIZE>>1)+1];
	long j=0,n=0,i=0,l=0;
	gostc_POINT64 P;
	int rc = -1;
	
	if(!params || !mas|| !p1)
	{
		rc = -1;
		goto end;
	}
	ZZ(k);
	ZZ(gamma);
	ZZ(pmas);
	ZZ(NZ);
	ZZS(P);
	memcpy(pmas,mas,ASYM_DEF_512_SIZE);
	
	for(j=0; j<ASYM_DEF_512_SIZE>>2; j++)
	{
		l = 0;
		while(pmas[j]>1)
		{
			k[j*32+l+1] = pmas[j]&1;
			pmas[j] = pmas[j]>>1;
			l++;
		}
		k[j*32+l+1] = pmas[j];	
	} /*//заполним к от 1 до 512*/
	k[0] = k[ASYM_BIT_512_SZ+1] = 0;
	
	n = ASYM_BIT_512_SZ;
	while(n > 0 && k[n]==0)
		n--;
	if(n==0) 
	{
		rc = -1;/*Вырожденный случай: "нулевая" кратная точка. В нашем случае ошибка*/
		goto end;
	}
	i = n;

	
	/*//step-1*/
	memcpy(&P,&params->PM0,ASYM_POINTSZ_512BIT*sizeof(uint32_t));
	
	i--;
			
	/*//step-2 */
	
	while(i >= 1)
	{
		if(gostc_DoubleInProj64M(params,&P,NULL)<0)
		{
			rc = -2;
			goto end;
		}
		if(k[i] == 1)
		{
			
			if(gostc_AddInProj64M(&P,(gostc_pPOINT64)&params->PM0,&P,params)<0)
			{
				rc = -3;
				goto end;
			}	
		}
		i--;
	}
	/*///Нахождение обратных чисел... (Step - 9)*/
	
	if(!noconvert)
	{/*//Выводим координаты из Монтгомери*/
		memcpy(NZ,P.Z,ASYM_DEF_512_SIZE);
		if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
		{
			rc = -2;
			goto end;
		}
		memcpy(P.Z,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
		ZZ(NZ);
		memcpy(NZ,P.X,ASYM_DEF_512_SIZE);
		if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
		{
			rc = -2;
			goto end;
		}
		memcpy(P.X,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
		ZZ(NZ);
		memcpy(NZ,P.Y,ASYM_DEF_512_SIZE);
		if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
		{
			rc = -2;
			goto end;
		}
		memcpy(P.Y,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
		/*//Гамма*/
		if(gostc_BinaryEuclid64(P.Z,params->P,gamma)<0)
		{
			rc = -4;
			goto end;
		}
		ZZ(NZ);

		if(gostc_Find16ModP(gamma,P.Z,params->P,NZ)<0)
		{
			rc=-5;
			goto end;
		}

		
	/*////////////////step 9///////////////////////////// */
		if(gostc_Find16ModP(gamma,P.X,params->P,P.X)<0)
			{rc =-5; goto end;}
		if(gostc_Find16ModP(gamma,P.X,params->P,p1->xp)<0)
			{rc =-5; goto end;}
		if(gostc_Find16ModP(gamma,P.Y,params->P,P.Y)<0)
			{rc =-5; goto end;}
		if(gostc_Find16ModP(gamma,P.Y,params->P,P.Y)<0)
			{rc =-5; goto end;}
		if(gostc_Find16ModP(gamma,P.Y,params->P,p1->yp)<0)
			{rc =-5; goto end;}
	}
	else
	{
		memcpy(p1->xp,P.X,ASYM_DEF_512_SIZE);
		memcpy(p1->yp,P.Y,ASYM_DEF_512_SIZE);
		memcpy(p1->Zp,P.Z,ASYM_DEF_512_SIZE);
	}

	rc = 0;/*OK*/
end:
	SZZ(k);
	SZZ(gamma);
	SZZ(pmas);
	SZZ(NZ);
	SZZS(j);
	SZZS(n);
	SZZS(i);
	SZZS(l);
	SZZS(P);

	return rc;
}

/*////////////////////////////////////////////////////////////////// */
/*Нахождение суммы кратных точек k1P + k2Q, QM0 - представление точки Q, mas1 и mas2 - представления чисел k1 и k2
Вычисленная x-координата cуммы сохраняются в x_result 
Возврат: 0 - ОК, <0 - ошибка 	*/

int gostc_FindSumOfKrPoints(const gostc_CR_SYSTEM* params, gostc_pONLY_POINT QM0, uint32_t *mas1, uint32_t *mas2,  uint32_t *x_result)
{
	uint32_t k1[ASYM_BIT_SZ>>1], k2[ASYM_BIT_SZ>>1];
	uint32_t pmas1[ASYM_DEF_SIZE>>2], pmas2[ASYM_DEF_SIZE>>2];
	int32_t j=0;
	int32_t l=0, n=0, i=0;
	gostc_ONLY_POINT P,pointADD[16];
	uint8_t zero[ASYM_DEF_SIZE];
	uint32_t NZ[(ASYM_DEF_SIZE>>1)+1];
	uint32_t gamma[ASYM_DEF_SIZE>>2];
	int rc = -1;

	if(!params || !QM0 || !mas1 || !mas2 || !x_result)/*params error*/
		return rc;

	ZZ(k1);
	ZZ(k2);
	ZZ(zero);
	ZZ(NZ);
	ZZ(gamma);
	ZZ(pointADD);
	ZZS(P);

	memcpy(pmas1,mas1, ASYM_DEF_SIZE);
	memcpy(pmas2,mas2, ASYM_DEF_SIZE);

	for (j = 0; j <  ASYM_DEF_SIZE>>2; j++) {
		l = 0;
		while(pmas1[j] > 3)
		{
			k1[j*16+l] = pmas1[j]&3;
			pmas1[j] = pmas1[j]>>2;
			l++;
		}
		k1[j*16+l] = pmas1[j];
	}
	for(j=0; j< ASYM_DEF_SIZE>>2; j++) {
		l = 0;
		while(pmas2[j] > 3)
		{
			k2[j*16+l] = pmas2[j]&3;
			pmas2[j] = pmas2[j]>>2;
			l++;
		}
		k2[j*16+l] = pmas2[j];	
	}
	n = ASYM_BIT_SZ>>1;
	while(n && (k1[n - 1] == 0) && (k2[n - 1] == 0))
		n--;
	if(n==0) { 
		rc = -1;
		goto end;
	}
/************************************************************************/
	/* Предрасчеты... */
/************************************************************************/
	memcpy(&pointADD[1],QM0,ASYM_POINTSZ_256BIT*sizeof(uint32_t));/* Q */
	if(gostc_DoubleInProjectiveWM(params, &pointADD[1], &pointADD[2])<0)/*2Q*/
	{
		rc = -2;
		goto end;
	}
	if(gostc_AddingInProjectiveWInM(&pointADD[2],&pointADD[1],&pointADD[3],params)<0)
	{
		rc = -3;
		goto end;
	}
	memcpy(&pointADD[4],&params->PM0,ASYM_POINTSZ_256BIT*sizeof(uint32_t));/*P*/

	for(j=1;j<4;j++)
	{/* P+Q,P+2Q,P+3Q */
		if(gostc_AddingInProjectiveWInM(&pointADD[j], &pointADD[4], &pointADD[j+4],params)<0)
		{
			rc = -3;
			goto end;
		}
	
	}
	memcpy(&pointADD[8],&params->P2M0,ASYM_POINTSZ_256BIT*sizeof(uint32_t));/*2P*/
		
	for(j=1;j<4;j++)
	{/* 2P+Q,2P+2Q,2P+3Q */
		if(gostc_AddingInProjectiveWInM(&pointADD[j], &pointADD[8], &pointADD[j+8],params)<0)
		{
			rc = -3;
			goto end;
		}
	}
	memcpy(&pointADD[12],&params->P3M0,ASYM_POINTSZ_256BIT*sizeof(uint32_t));/*3P*/
	for(j=1;j<4;j++)
	{/*3P+Q,3P+2Q,3P+3Q*/
		if(gostc_AddingInProjectiveWInM(&pointADD[j], &pointADD[12], &pointADD[j+12], params)<0)
		{
			rc = -3;
			goto end;
		}
	}
/*******************Поиск суммы кратных точек********************************/
	/* step-2 */
	i = n;
	/* step-3 */
SumS3:
	if(memcmp(P.Z,zero,ASYM_DEF_SIZE)==0)
		goto SumS5;
/* "двойное" Удвоение... */
	if(gostc_DoubleInProjectiveWM(params,&P,NULL)<0)
	{
		rc = -2;	
		goto end;
	}
	if(gostc_DoubleInProjectiveWM(params,&P,NULL)<0)
	{
		rc = -2;	
		goto end;
	}
SumS5:
	if ((k1[i-1] == 0)&&(k2[i-1] ==0)) goto SumS7;
	if(memcmp(P.Z,zero,ASYM_DEF_SIZE)==0) /* т.к Z=0, то */
	{
		memcpy(&P,&pointADD[k1[i-1]*4+k2[i-1]],ASYM_POINTSZ_256BIT*sizeof(uint32_t));
		goto SumS7;		
	}
	if(gostc_AddingInProjectiveWInM(&P,&pointADD[k1[i-1]*4+k2[i-1]],&P,params)<0)
	{
		rc = -3;
		goto end;
	}
	
	
SumS7:
	i--;
	if(i != 0)
		goto SumS3;
	memcpy(NZ,P.Z, ASYM_DEF_SIZE);

	if(gostc_ModP_C(NZ,params->P,params->EpsP,8)<0)
	{
		rc = -3;
		goto end;
	}
	memcpy(P.Z,NZ+(ASYM_DEF_SIZE>>2), ASYM_DEF_SIZE);
		
	ZZ(NZ);

	memcpy(NZ,P.X, ASYM_DEF_SIZE);
	if(gostc_ModP_C(NZ,params->P,params->EpsP,8)<0)
	{
		rc = -3;
		goto end;
	}
	memcpy(P.X,NZ+(ASYM_DEF_SIZE>>2), ASYM_DEF_SIZE);
	
/*	BinaryEuclid(P.Z,params->P,gamma,&params->p1); */
	if(gostc_BinaryEuclid_C(P.Z,params->P,gamma)<0)
	{
		rc = -4;
		goto end;
	}

/* 	inverce(P.Z+2,params->P,gamma2+2,&params->p1); */
	if(gostc_FindEQABModP(gamma,P.X,params->P,P.X)<0)
		{rc =-5; goto end;}
	if(gostc_FindEQABModP(gamma,P.X,params->P,x_result)<0)
		{rc =-5; goto end;}
		
	
	rc = 0;
end:
	SZZ(k1);
	SZZ(k2);
	SZZ(pmas1);
	SZZ(pmas2);
/*	ZZ(zero); */
	SZZ(NZ);
	SZZ(gamma);
	SZZ(pointADD);
	SZZS(P);
	return rc;
}
/*************************************************************************************************/
int gostc_FindSumKrPoints64(const gostc_CR_SYSTEM64* params, gostc_pPOINT64 QM0, uint8_t *mas1, uint8_t *mas2,uint32_t *x_result)
{
	uint32_t k1[ASYM_BIT_512_SZ+2],k2[ASYM_BIT_512_SZ+2];
	uint32_t koef=0;
	uint32_t pmas1[ASYM_DEF_512_SIZE>>2],pmas2[ASYM_DEF_512_SIZE>>2],NZ[(ASYM_DEF_512_SIZE>>1)+1],gamma[ASYM_DEF_512_SIZE>>2];
	long n=0,i=0,j=0,l=0;
	gostc_POINT64 P, PplusQ;
	int rc = -1,ret=0;
	
	if(!params || !QM0 || !mas1 || !mas2 || !x_result)
	{
		rc = -1;
		return rc;
	}

	ZZ(k1);
	ZZ(k2);
	ZZS(P);
	ZZS(PplusQ);

	memcpy(pmas1,mas1,ASYM_DEF_512_SIZE);
	memcpy(pmas2,mas2,ASYM_DEF_512_SIZE);
	l = 0;
	for(j=0; j<ASYM_DEF_512_SIZE>>2; j++)
	{
		l = 0;	
		while(pmas1[j]>1)
		{
			k1[j*32+l+1] = pmas1[j]&1;
			pmas1[j] = pmas1[j]>>1;
			l++;
		}
		k1[j*32+l+1] = pmas1[j];	
	}/*//заполним k1 и k2 от 1 до 512*/
	k1[0] = k1[ASYM_BIT_512_SZ+1] = k2[0] = k2[ASYM_BIT_512_SZ+1] = 0;

	for(j=0; j<ASYM_DEF_512_SIZE>>2; j++)
	{
		l = 0;
		while(pmas2[j] > 1)
		{
			k2[j*32+l+1] = pmas2[j]&1;
			pmas2[j] = pmas2[j]/2;
			l++;
		}
		k2[j*32+l+1] = pmas2[j];	
	}
	
	n = ASYM_BIT_512_SZ;

	while(n>0 && (k1[n]==0)&&(k2[n]==0))
		n--;
	if(n==0) 
	{
		rc = -2;
		goto end; /*////Ошибочка!!!*/
	}

	

	if(gostc_AddInProj64M((gostc_pPOINT64)&params->PM0,QM0,&PplusQ,params)<0)
	{
		rc = -3;
		goto end;
	}
		
	/*//step-2 */
	i = n;
	/*//step-3 Кладем стартовую точку*/
	koef = k2[i]*2 + k1[i];
	if(koef == 1)
		memcpy(&P,&params->PM0,ASYM_POINTSZ_512BIT*sizeof(uint32_t));
	if(koef == 2)
		memcpy(&P,QM0,ASYM_POINTSZ_512BIT*sizeof(uint32_t));
	if(koef == 3)
		memcpy(&P,&PplusQ,ASYM_POINTSZ_512BIT*sizeof(uint32_t));

	i--;
	while(i>0)
	{
/*//Удвоение...*/
		if(gostc_DoubleInProj64M(params,&P,NULL)<0)
		{
			rc = -2;	
			goto end;
		}
		
		if ((k1[i] == 0)&&(k2[i] ==0))
		{

		}
		else
		{
			koef = k2[i]*2 + k1[i];
			
			ret =0;
			if(koef == 1)
				ret =gostc_AddInProj64M(&P,(gostc_pPOINT64)&params->PM0,&P,params);
			else if(koef == 2)
				ret = gostc_AddInProj64M(&P,QM0,&P,params);
			else if(koef == 3)
				ret =gostc_AddInProj64M(&P,&PplusQ,&P,params);
			if(ret < 0)
			{
				rc = -2;
				goto end;
			}
		}
		i--;
	}
/*	////Выводим из представления Монтгомери в обычное*/
	ZZ(NZ);
	memcpy(NZ,P.Z,ASYM_DEF_512_SIZE);
	if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc = -3;
		goto end;
	}

	memcpy(P.Z,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
	
	ZZ(NZ);
	memcpy(NZ,P.X,ASYM_DEF_512_SIZE);
	if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc = -3;
		goto end;
	}

	memcpy(P.X,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);

	ZZ(gamma);
	if(gostc_BinaryEuclid64(P.Z,params->P,gamma)<0)
	{
		rc = -4;
		goto end;
	}

	if(gostc_Find16ModP(gamma,P.X,params->P,P.X)<0)
		{rc =-5; goto end;}
	if(gostc_Find16ModP(gamma,P.X,params->P,x_result)<0)
		{rc =-5; goto end;}

	rc = 0;/*OK*/

end:
	SZZ(k1);
	SZZ(k2);
	SZZ(pmas1);
	SZZ(pmas2);
	SZZ(NZ);
	SZZ(gamma);
	SZZS(koef);
	SZZS(n);
	SZZS(i);
	SZZS(j);
	SZZS(l);
	SZZS(P);
	SZZS(PplusQ);

	return rc;
}

/*Нахождение кратной точки по алгоритму Диффи-Хэллмана: mas*pbase*/
/*Результат сохраняется в prez					*/
/*Возврат: 0 - ОК, <0 - ошибка 	*/
int gostc_FindKrPointDiffiHellman(const gostc_CR_SYSTEM *params, uint8_t *mas,gostc_pONLY_POINT pbase, gostc_POINT_ECP *prez)
{
	uint8_t k[ASYM_BIT_SZ];
	uint8_t pmas[ASYM_DEF_SIZE];
	int i=0, j=0, l=0, n=0;
	gostc_ONLY_POINT P;
	uint8_t zero[ASYM_DEF_SIZE];
	uint32_t NZ[(ASYM_DEF_SIZE>>1)+2];
	uint32_t gamma[ASYM_DEF_SIZE>>2];
	int rc = -1;

	if(!params || !mas || !pbase || !prez)
		return rc;/*Ошибка передачи параметров!*/

	ZZ(k);
/*	ZZ(pmas); */
	ZZ(zero);
	ZZ(NZ);
	ZZ(gamma);
	ZZS(P);

	memcpy(pmas,mas,ASYM_DEF_SIZE);
	for(j=0; j<ASYM_DEF_SIZE; j++)
	{
		l = 0;
		while(pmas[j] > 1)
		{
			k[j*8+l] = pmas[j]&1;
			pmas[j] = pmas[j]>>1;
			l++;
		}
		k[j*8+l] = pmas[j];
	}
	n = ASYM_BIT_SZ;
	while(n >0 && k[n-1]==0)
		n--;
	if(n==0) 
	{
		rc = -1;/*Вырожденный случай: нулевая кратная точка. В нашей ситуации это ошибка!*/
		goto end;
	}
	P.X[0] = P.Y[0] = 1;
	i = n - 1;
RepeatOnStep3:
	/*Step3:*/
	/*пусть K = 2 (10)k0 = 0, k1 = 1;*/
	
	if(memcmp(P.Z,zero,ASYM_DEF_SIZE)==0)
		goto OnStep5;
	
	/*Удвоение*/
	if(gostc_DoubleInProjectiveWM(params,&P,NULL)<0)
	{
		rc = -2;	
		goto end;
	}
		
OnStep5:
/*///////////////////////////step - 5////////////////////////////////// */
	if (k[i] == 0) goto OnStep7;
	if(memcmp(P.Z,zero,ASYM_DEF_SIZE)==0)/*т.к Z=0, то*/
	{
		memcpy(&P,pbase,ASYM_POINTSZ_256BIT*sizeof(uint32_t));
		goto OnStep7;
	}
	if(gostc_AddingInProjectiveWInM(&P,pbase,&P,params)<0)
	{
		rc = -2;	
		goto end;
	}
		
OnStep7:
	i--;
	if (i >=0) /* (Step - 8) */
		goto RepeatOnStep3;
	/* Нахождение обратных чисел... (Step - 9) */
	
	/* Выводим координаты из Монтгомери */
	memcpy(NZ,P.Z,ASYM_DEF_SIZE);
	if(gostc_ModP_C(NZ,params->P,params->EpsP,(ASYM_DEF_SIZE>>2))<0)
	{
		rc = -3;	
		goto end;
	}
	memcpy(P.Z,NZ+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
	ZZ(NZ);
	
	memcpy(NZ,P.X,ASYM_DEF_SIZE);
	if(gostc_ModP_C(NZ,params->P,params->EpsP,(ASYM_DEF_SIZE>>2))<0)
	{
		rc = -3;	
		goto end;
	}
	memcpy(P.X,NZ+(ASYM_DEF_SIZE>>2),ASYM_DEF_SIZE);
	ZZ(NZ);
	memcpy(NZ,P.Y,ASYM_DEF_SIZE);
	if(gostc_ModP_C(NZ,params->P,params->EpsP,(ASYM_DEF_SIZE>>2))<0)
	{
		rc = -3;	
		goto end;
	}
	memcpy(P.Y,NZ+(ASYM_DEF_SIZE>>2), ASYM_DEF_SIZE);
	/* Гамма */
	if(gostc_BinaryEuclid_C(P.Z,params->P,gamma)<0)
	{
		rc = -4;
		goto end;
	}
	/*///////////////step 9///////////////////////////// */
	if(gostc_FindEQABModP(gamma,P.X,params->P,P.X)<0)
		{rc =-5; goto end;}	
	if(gostc_FindEQABModP(gamma,P.X,params->P,prez->xp)<0)
		{rc =-5; goto end;}	
	if(gostc_FindEQABModP(gamma,P.Y,params->P,P.Y)<0)
		{rc =-5; goto end;}
	if(gostc_FindEQABModP(gamma,P.Y,params->P,P.Y)<0)
		{rc =-5; goto end;}
	if(gostc_FindEQABModP(gamma,P.Y,params->P,prez->yp)<0)
		{rc =-5; goto end;}
	
	rc = 0;/*ОК*/
end:
	SZZ(k);
	SZZ(pmas);
	SZZ(NZ);
	SZZ(gamma);
	SZZS(P);
	return rc;
}
/*********************для 512-битных наборов******************************************************/
int gostc_FindKrPointDiffiHellman64(const gostc_CR_SYSTEM64*  params, uint8_t* mas, gostc_pPOINT64 pbase,gostc_POINT_ECP64 *prez)
{
	int k[ASYM_BIT_512_SZ+2];
	uint32_t pmas[ASYM_DEF_512_SIZE>>2];
	int j=0,i=0,n=0,l=0;
	gostc_POINT64 P;
	uint32_t NZ[(ASYM_DEF_512_SIZE>>1)+1],gamma[ASYM_DEF_512_SIZE>>2];
	int rc = -1;

	if(!params || !mas || !pbase || !prez)
		return rc;/*Ошибка передачи параметров!*/

	ZZS(P);
	ZZ(k);
	ZZ(gamma);
	ZZ(pmas);
	ZZ(NZ);
	
	memcpy(pmas,mas,ASYM_DEF_512_SIZE);
	
	for(j=0; j<ASYM_DEF_512_SIZE>>2; j++)
	{
		l = 0;
		while(pmas[j]>1)
		{
			k[j*32+l+1] = pmas[j]&1;
			pmas[j] = pmas[j]>>1;
			l++;
		}
		k[j*32+l+1] = pmas[j];	
	} /*//заполним к от 1 до 512*/
	k[0] = k[ASYM_BIT_512_SZ+1] = 0;
	
	
	n = ASYM_BIT_512_SZ;
	while(n>0 && k[n]==0)
		n--;
	if(n==0)
	{
		rc = -2;/*Вырожденный случай: нулевая кратная точка. В нашей ситуации это ошибка!*/
		goto end;
	}
			/*//step-2 */
	i = n;
	memcpy(&P,pbase,ASYM_POINTSZ_512BIT*sizeof(uint32_t));/*/////Базовая точка - чужой открытый ключ*/
	i--;
	/*//Step3:*/
	while(i>0)
	{
		/*//Удвоение*/
		if(gostc_DoubleInProj64M(params,&P,NULL)<0)
		{
			rc = -2;	
			goto end;
		}
		if (k[i] != 0) 
		{/*/////Сложение*/
			if(gostc_AddInProj64M(&P,pbase,&P,params)<0)
			{
				rc = -2;
				goto end;

			}
		}
		i--;
	}
		/*///Нахождение обратных чисел... (Step - 9)*/
	/*	//Выводим координаты из Монтгомери*/
	memcpy(NZ,P.Z,ASYM_DEF_512_SIZE);
	if(gostc_ModP16(NZ,params->P,params->EpsP,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc = -3;
		goto end;
	}

	memcpy(P.Z,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
	ZZ(NZ);
	memcpy(NZ,P.X,ASYM_DEF_512_SIZE);
	if(gostc_ModP16(NZ,params->P,params->EpsP,(ASYM_DEF_512_SIZE>>2))<0)
	{
		rc = -3;
		goto end;
	}

	memcpy(P.X,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
	ZZ(NZ);
	memcpy(NZ,P.Y,ASYM_DEF_512_SIZE);
	if(gostc_ModP16(NZ,params->P,params->EpsP,(ASYM_DEF_512_SIZE>>2))<0)
	{
		rc = -3;
		goto end;
	}
	memcpy(P.Y,NZ+(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
	if(gostc_BinaryEuclid64(P.Z,params->P,gamma)<0)
	{
		rc = -4;
		goto end;
	}
	
	if(gostc_Find16ModP(gamma,P.X,params->P,P.X)<0)
		{rc=-5; goto end;}
	if(gostc_Find16ModP(gamma,P.X,params->P,prez->xp)<0)
		{rc=-5; goto end;}
	if(gostc_Find16ModP(gamma,P.Y,params->P,P.Y)<0)
		{rc=-5; goto end;}
	if(gostc_Find16ModP(gamma,P.Y,params->P,P.Y)<0)
		{rc=-5; goto end;}
	if(gostc_Find16ModP(gamma,P.Y,params->P,prez->yp)<0)
		{rc=-5; goto end;}
	
	rc = 0;/*OK*/
end:
	SZZ(k);
	SZZ(pmas);
	SZZS(j);
	SZZS(i);
	SZZS(n);
	SZZS(l);
	SZZS(P);
	SZZ(NZ);
	SZZ(gamma);
	
	return rc;
}
/************************************************************/



/*////////////Маскирование*/

static const uint32_t Q0[] = 
{
	0x87e6e6df,0xd644aaf1,0xbe23c595,0xd86e25ed,0xecc423f1,0x19905c5e,0xcb1dbac7,0xa82f2d7e,
	0xb5d2d15d,0xd4eb7c09,0xffb90f04,0x922b14b2,0x6b2fee80,0x550d267b,0xfe0023c7,0x4531acd1,/*Q - gost2012-test-set*/
	0x1F10B275, 0xCACDB141,0xFAD2B85D,0x9B4B38AB,0x4E056060,0x6FF22B8D,0xF48D8911,0x27E69532,/*Q - gost2012ParamSetA*/
	0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,	
	0x374F25BD, 0xC6346C54, 0x101BEA0E,0x8B996712,0xD9D40CFA,0xACFDB77B,0x2565A545,0x49A1EC14,/*Q - gost2012ParamSetB*/
	1,0,0,0,0,0,0, ASYM_ELDER,
};

int  gostc_DoMask64(void *key, void* mask,uint32_t n) 
{
	uint32_t QM[(ASYM_DEF_512_SIZE>>2)+1];
	uint32_t k[(ASYM_DEF_512_SIZE>>2)+1];
	uint32_t m[(ASYM_DEF_512_SIZE>>2)+1],zero[ASYM_DEF_512_SIZE>>2],buf[34],buf2[(ASYM_DEF_512_SIZE>>2)+2];
	int res_cmp=0;
	int rc = 0;

	if(!key || !mask || n<MASK_TST512 || n > MASK_TC26B)
		return 0;	/*Ошибка передачи параметров*/
		
	
	ZZ(k);
	ZZ(m);
	ZZ(zero);
	ZZ(buf);
	ZZ(buf2);
	ZZ(QM);

	memcpy(k,key,ASYM_DEF_512_SIZE);
	memcpy(m,mask,ASYM_DEF_512_SIZE);
	memcpy(QM,Q0+(n-MASK_TST512)*(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);

	/*///mask = mask(mod Q)*/
	
	if(gostc_Divide_c16(m,QM,buf,ASYM_DEF_512_SIZE>>2) == 0) goto end;

	res_cmp = gostc_Comp_C(m,zero,ASYM_DEF_512_SIZE>>2);
	if(res_cmp <0)
		goto end;/*Ошибка*/	
	if(res_cmp == 2)
		m[0] = 1;
	/*///buf = key * mask;*/
	if(gostc_Mul16(k,m,buf,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc =0;
		goto end;
	}
	if(gostc_Divide_c16(buf,QM,buf2,ASYM_DEF_512_SIZE>>1) == 0) goto end;

	memcpy(key,buf,ASYM_DEF_512_SIZE);

	rc = 1;
end:
	SZZ(QM);
	SZZ(k);
	SZZ(m);
	SZZ(buf);
	SZZ(buf2);
	SZZS(res_cmp);

	return rc;
}

/*////////////////////Снятие маски*/

int gostc_DoUnmask64(void *key,void*mask,uint32_t n)
{

	uint32_t Q[(ASYM_DEF_512_SIZE>>2)+1];/*//Q - ГОСТ2012-set*/
	uint32_t k[(ASYM_DEF_512_SIZE>>2)+1], m[(ASYM_DEF_512_SIZE>>2)+1],zero[ASYM_DEF_512_SIZE>>2],buf[(ASYM_DEF_512_SIZE>>1)+1],buf2[(ASYM_DEF_512_SIZE>>2)+1],
gamma[ASYM_DEF_512_SIZE>>2];
	int res_cmp=0;
	int rc = 0;
	if(!key || !mask || n<MASK_TST512 || n > MASK_TC26B)
		return 0;	/*Ошибка передачи параметров*/

	ZZ(m);	
	ZZ(zero);
	ZZ(buf);
	ZZ(buf2);
	ZZ(gamma);
	ZZ(Q);

	memcpy(Q,Q0+(n-MASK_TST512)*(ASYM_DEF_512_SIZE>>2),ASYM_DEF_512_SIZE);
		
	memcpy(k,key,ASYM_DEF_512_SIZE);
	memcpy(m,mask,ASYM_DEF_512_SIZE);
	
	
	
	if(gostc_Divide_c16(m,Q,buf,ASYM_DEF_512_SIZE>>2) == 0) goto end;

	res_cmp = gostc_Comp_C(m,zero,ASYM_DEF_512_SIZE>>2);
	if(res_cmp < 0)
		goto end;/*ОШибка*/	
	if(res_cmp == 2)
		m[0] = 1;

	if(gostc_BinaryEuclid64(m,Q,gamma)<0)
	{
		rc = 0;
		goto end;
	}
	if(gostc_Mul16(k,gamma,buf,ASYM_DEF_512_SIZE>>2)<0)
	{
		rc =0;
		goto end;
	}
	
	if(gostc_Divide_c16(buf,Q,buf2,ASYM_DEF_512_SIZE>>1) == 0) goto end;
	memcpy(key,buf,ASYM_DEF_512_SIZE);

	rc = 1;
end:
	SZZ(Q);
	SZZ(k);
	SZZ(m);
	SZZ(buf);
	SZZ(buf2);
	SZZ(gamma);
	SZZS(res_cmp);
	
	return rc;
}


/*Процедура вычисления ключа ВКО (256-битные наборы)
Возврат: 0 - ошибка, 1 - ОК!!! */
int gostc_GenDFHKey32(const gostc_CR_SYSTEM *par,void* keyD,void * ukm16, void* keyQ,void* keyDFH)
{
	uint32_t zero[ASYM_DEF_SIZE>>2], local_UKM[ASYM_DEF_SIZE>>2];
	uint8_t kkey[ASYM_DEF_SIZE];
	int a=-1;
	gostc_ONLY_POINT pQ;
	gostc_POINT_ECP p1;
	int rc = 0;

	if(!par || !keyD || ! ukm16 || !keyQ || !keyDFH)
		return rc;/*Ошибка передачи параметров*/

	ZZ(zero);
	ZZ(kkey);
	ZZS(pQ);
	ZZS(p1);
	ZZ(local_UKM);
	a = gostc_Comp_C(zero,keyQ,ASYM_DEF_SIZE>>2);
	if (a == 2|| a <0) 
		return rc;/*rc=0*/
	a = gostc_Comp_C(keyD,zero,ASYM_DEF_SIZE>>2);
	if(a == 2|| a <0) /*rc=0*/
		return rc;
	a = gostc_Comp_C(ukm16, zero, ASYM_UKM16_SIZE>>2);
	if(a == 2 || a <0) /*rc=0*/
		return rc;
	
	memcpy(kkey,keyD,ASYM_DEF_SIZE);
	memcpy(pQ.X,keyQ,ASYM_DEF_SIZE);
	memcpy(pQ.Y,(uint8_t*)keyQ+ASYM_DEF_SIZE,ASYM_DEF_SIZE);
	memcpy(local_UKM, ukm16, ASYM_UKM16_SIZE); /* Сохранили UKM */
	/* Рассчитаем UKM*kkey modQ */
	
	if(gostc_FindEQABModP((uint32_t*)(kkey),local_UKM,par->Q,(uint32_t*)(kkey))<0)
		goto end;/*rc=0*/

	pQ.Z[0] = 1; /* Заполнили значение точки (откр. ключ) в проективных коорд. */
	/* теперь переведем её в Монтгомери */
	if(gostc_FindEQABModPinM(par,pQ.X,par->R,par->P,pQ.X)<0)
		goto end;/*rc=0*/

	if(gostc_FindEQABModPinM(par,pQ.Y,par->R,par->P,pQ.Y)<0)
		goto end;/*rc=0*/
	
	if(gostc_FindEQABModPinM(par,pQ.Z,par->R,par->P,pQ.Z)<0)
		goto end;/*rc=0*/
	
	/* Кратная точка Диффи-Хеллмана */
	if(gostc_FindKrPointDiffiHellman(par,kkey,&pQ, &p1)<0)
		goto end;/*rc=0*/
	
	memcpy(keyDFH,p1.xp,ASYM_DEF_SIZE);
	memcpy((uint8_t*)keyDFH+ASYM_DEF_SIZE,p1.yp,ASYM_DEF_SIZE);
	rc = 1;
end:
	SZZS(p1);
	SZZ(kkey);
	SZZS(pQ);
	SZZ(local_UKM);

	return rc;
}
/*Процедура вычисления ключа ВКО (512-битные наборы)
Возврат: 0 - ошибка, 1 - ОК!!! */
int gostc_GenDFHKey64(const gostc_CR_SYSTEM64* par,void* keyD,void * ukm16, void* keyQ,void* keyDFH)
{
	uint32_t zero[ASYM_DEF_512_SIZE>>2],kD[ASYM_DEF_512_SIZE>>2],svL[ASYM_DEF_512_SIZE>>2],Q[(ASYM_DEF_512_SIZE>>2)+1];
	uint8_t kkey[ASYM_DEF_512_SIZE];
	gostc_POINT64 pQ;
	gostc_POINT_ECP64 p1;
	int a=-1,rc=0;
	
	if(!par || !keyD || ! ukm16 || !keyQ || !keyDFH)
		return rc;/*Ошибка передачи параметров*/


	ZZ(zero);
	ZZ(kD);
	ZZ(svL);
	ZZ(Q);
	ZZ(kkey);
	ZZS(pQ);
	ZZS(p1);

	a = gostc_Comp_C(zero,keyQ,ASYM_DEF_512_SIZE>>1);
	if (a == 2 || a <0) 
		return rc;
	a = gostc_Comp_C(keyD,zero,ASYM_DEF_512_SIZE>>2);
	if(a == 2 || a <0)
		return rc;

	a = gostc_Comp_C(ukm16,zero, ASYM_UKM16_SIZE>>2);
	if(a == 2 || a <0)
		return rc;

	/*/////Умножим keyD на sv по модулю Q*/
	
/*	f1 = fopen("2.txt","a");*/
	memcpy(svL, ukm16, ASYM_DEF_512_SIZE>>2);
	memcpy(kD,keyD,ASYM_DEF_512_SIZE);
	memcpy(Q,par->Q,ASYM_DEF_512_SIZE);

	
	if(gostc_Find16ModP(svL,kD,Q,kD)<0)
		goto end;

	memcpy(kkey,kD,ASYM_DEF_512_SIZE); /*//А это сейчас.*/


	memcpy(pQ.X,keyQ,ASYM_DEF_512_SIZE);
	memcpy(pQ.Y,(uint8_t*)keyQ+ASYM_DEF_512_SIZE,ASYM_DEF_512_SIZE);
	/*//pQ.Z[0] = 1;//Заполнили значение точки (откр. ключ) в проективных коорд.*/
	/*//теперь переведем её в Монтгомери*/
	if(gostc_Find16ModPinM(par,pQ.X,par->R,par->P,pQ.X)<0)
		goto end;
	
	if(gostc_Find16ModPinM(par,pQ.Y,par->R,par->P,pQ.Y)<0)
		goto end;
	
	memcpy(pQ.Z,par->M1,ASYM_DEF_512_SIZE);
	
	if(gostc_FindKrPointDiffiHellman64(par,kkey,&pQ,&p1)<0)
		goto end;
	memcpy(keyDFH,p1.xp,ASYM_DEF_512_SIZE);
	memcpy((uint8_t*)keyDFH+ASYM_DEF_512_SIZE,p1.yp,ASYM_DEF_512_SIZE);
	rc=1; /*OK*/
end:
	SZZS(p1);
	SZZ(kD);
	SZZ(svL);
	SZZ(Q);
	SZZ(kkey);
	SZZS(pQ);
	SZZS(a);
	

	return rc;
}
/**********************************************************************************/




