#include "gostc_mem.h"
#include "gostc_sym.h"

#define ZZ(x) bzero(x, sizeof(x))
#define ZZS(x) bzero(&x, sizeof(x))
#define SZZ(x) gostc_cleanse(x, sizeof(x))
#define SZZS(x) gostc_cleanse(&x, sizeof(x))



#define rotl32(x,n)   (((x) << ((uint16_t)(n))) | ((x) >> (32 - (uint16_t)(n))))


 #define cycle( half1, half2, key, mask,ltab )                       \
                tmp1 = (key^mask) + half1;                             \
                tmp2 = half2 ^ ltab[tmp1 & 0xff];            \
                tmp2 ^= ltab[256+((tmp1 >>= 8) & 0xff)];           \
                half2 = half1;                                  \
                tmp2 ^= ltab[512+((tmp1 >>= 8) & 0xff)];           \
                tmp2 ^= ltab[768+((tmp1 >>= 8) & 0xff)];           \
                half1 = tmp2;



void gostc_sym_ECB( uint32_t *InBuffer, uint32_t *OutBuffer, uint32_t *key, 
	uint32_t *mask, uint32_t* lbox)
{
        register uint32_t a = InBuffer[1];
        register uint32_t b = InBuffer[0];
	register uint32_t tmp1,tmp2; 

        cycle( b, a, key[0], mask[0],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[7], mask[7],lbox );

        cycle( b, a, key[0], mask[0],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[7], mask[7],lbox );

        cycle( b, a, key[0], mask[0],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[7], mask[7],lbox );

        cycle( b, a, key[7], mask[7],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[0], mask[0],lbox );

        ((uint32_t *) OutBuffer)[1] = b;
        ((uint32_t *) OutBuffer)[0] = a;
 }

void gostc_sym_ECBBack(uint32_t *InBuffer,uint32_t *OutBuffer, uint32_t *key, 
	uint32_t *mask,uint32_t* lbox)
{
        register uint32_t a = InBuffer[1];
        register uint32_t b = InBuffer[0];
        register uint32_t tmp1,tmp2; 

        cycle( b, a, key[0], mask[0],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[7], mask[7],lbox );

        cycle( b, a, key[7], mask[7],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[0], mask[0],lbox );

        cycle( b, a, key[7], mask[7],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[0], mask[0],lbox );

        cycle( b, a, key[7], mask[7],lbox );
        cycle( b, a, key[6], mask[6],lbox );
        cycle( b, a, key[5], mask[5],lbox );
        cycle( b, a, key[4], mask[4],lbox );
        cycle( b, a, key[3], mask[3],lbox );
        cycle( b, a, key[2], mask[2],lbox );
        cycle( b, a, key[1], mask[1],lbox );
        cycle( b, a, key[0], mask[0],lbox );

        ((uint32_t *) OutBuffer)[1] = b;
        ((uint32_t *) OutBuffer)[0] = a;
}


void gostc_sym_C16( uint32_t* A, uint32_t* B,uint32_t *key, uint32_t *mask, uint32_t* lbox)
{
	register uint32_t a = *A;
	register uint32_t b = *B;
	register uint32_t tmp1,tmp2;

       cycle( b, a, key[0], mask[0],lbox );
       cycle( b, a, key[1], mask[1],lbox );
       cycle( b, a, key[2], mask[2],lbox );
       cycle( b, a, key[3], mask[3],lbox );
       cycle( b, a, key[4], mask[4],lbox );
       cycle( b, a, key[5], mask[5],lbox );
       cycle( b, a, key[6], mask[6],lbox );
       cycle( b, a, key[7], mask[7],lbox );

       cycle( b, a, key[0], mask[0],lbox );
       cycle( b, a, key[1], mask[1],lbox );
       cycle( b, a, key[2], mask[2],lbox );
       cycle( b, a, key[3], mask[3],lbox );
       cycle( b, a, key[4], mask[4],lbox );
       cycle( b, a, key[5], mask[5],lbox );
       cycle( b, a, key[6], mask[6],lbox );
       cycle( b, a, key[7], mask[7],lbox );

       *A = a;
       *B = b;
}

static void Lrotl(uint32_t * what,uint16_t count)
{
	uint32_t tmp = rotl32(*what,count);
	*what = tmp;
}

static void ModC(uint32_t* what)
{
 uint32_t tmp;
 what[0]+=0x01010101L;
 tmp = what[1]+0x01010104L;
 if(tmp<what[1])
  tmp++;

 what[1]=tmp;

}

void gostc_sym_ModC(void* data)
{
	ModC((uint32_t*)data);
}

void gostc_sym_EncryptCFB(const uint8_t* pPlain,uint8_t* pCrypto,const void* pSync,
                          uint32_t nSize,uint32_t* k1,uint32_t* k1m,uint32_t* CurUZ)
{

        uint8_t Synchro[8];
        uint8_t gamma[8];
        int cycles;
        int i =0;
        int j = 0;
        int rest;

	if (!pPlain || !pCrypto || !pSync || !k1 || !k1m || !CurUZ) return;
        memcpy(Synchro,pSync,8);
        cycles = nSize/8;
        for(i =0;i<cycles;i++)
        {
                gostc_sym_ECB((uint32_t*)Synchro,(uint32_t*)gamma,k1,k1m,CurUZ);
                for(j = 0;j<8;j++)
                 Synchro[j] = *pCrypto++ = *pPlain++ ^ gamma[j];
        }

        rest = nSize%8;
        if(rest)
        {
                gostc_sym_ECB((uint32_t*)Synchro,(uint32_t*)gamma,(uint32_t*)k1,(uint32_t*)k1m,(uint32_t*)CurUZ);
                for(j = 0;j<rest;j++)
                 Synchro[j] = *pCrypto++ = *pPlain++ ^ gamma[j];
        }

	SZZ(gamma);
	SZZ(Synchro);
}

void gostc_sym_DecryptCFB(const uint8_t* pCrypto,uint8_t* pPlain,const void* pSync,
                 uint32_t nSize,uint32_t* k1,uint32_t* k1m,uint32_t* CurUZ)
{
        uint8_t gamma[8];
        uint8_t Synchro[8];
        int cycles;
        int i =0;
        int j = 0;
        int rest;

	if (!pPlain || !pCrypto || !pSync || !k1 || !k1m || !CurUZ) return;
        memcpy(Synchro,pSync,sizeof(Synchro));

        cycles = nSize/8;
        for(i =0;i<cycles;i++)
        {
                gostc_sym_ECB((uint32_t*)Synchro,(uint32_t*)gamma,k1,k1m,CurUZ);
                memcpy(Synchro,pCrypto,sizeof(Synchro));
                for(j = 0;j<8;j++)
                *pPlain++ = *pCrypto++ ^ gamma[j];
        }

        rest = nSize%8;
        if(rest)
        {
                gostc_sym_ECB((uint32_t*)Synchro,(uint32_t*)gamma,k1,k1m,CurUZ);
                for(j = 0;j<rest;j++)
                 *pPlain++ = *pCrypto++ ^ gamma[j];
        }
	SZZ(Synchro);
	SZZ(gamma);
}

void gostc_sym_Gamma(uint8_t * pIn,uint8_t * pOut,void * pSync,uint32_t nSize,
      uint32_t* k1,uint32_t* k1m,uint32_t* CurUZ)
{
        uint32_t Synchro[2];
        uint8_t gamma[8];
        int cycles;
        int i,j;
        int rest;


	if (!pIn || !pOut || !pSync || !k1 || !k1m || !CurUZ) return;

        gostc_sym_ECB((uint32_t*)pSync,Synchro,k1,k1m,CurUZ);

        cycles = nSize/8;
        for(i = 0;i<cycles; i++)
        {

                ModC(Synchro);
                gostc_sym_ECB(Synchro,(uint32_t*)gamma,k1,k1m,CurUZ);
                for(j = 0;j<8;j++)
                 *pOut++ = *pIn++ ^ gamma[j];
        }
        rest = nSize%8;
        if(rest)
        {
                ModC(Synchro);
                gostc_sym_ECB(Synchro,(uint32_t*)gamma,k1,k1m,CurUZ);
                for(j = 0;j<rest;j++)
                 *pOut++ = *pIn++ ^ gamma[j];
        }
	SZZ(Synchro);
	SZZ(gamma);
}



void gostc_sym_setUZ(uint8_t * shortUZ,uint32_t * longUZ)
{
        uint32_t oneElement;
        int16_t i,j;

	if (!shortUZ || !longUZ) return;

        for(i = 0;i<16;i++)
        {
                for(j = 0;j<16;j++)
                {
                        oneElement = (shortUZ[i*4]&0xf0)|(shortUZ[j*4]&0xf);
                        Lrotl(&oneElement,11);
                        longUZ[i*16+j] = oneElement;

                        oneElement = (shortUZ[i*4+1]&0xf0)|(shortUZ[j*4+1]&0xf);
                        Lrotl(&oneElement,11+8);
                        longUZ[256+i*16+j] = oneElement;

                        oneElement = (shortUZ[i*4+2]&0xf0)|(shortUZ[j*4+2]&0xf);
                        Lrotl(&oneElement,11+16);
                        longUZ[512+i*16+j] = oneElement;

                        oneElement = (shortUZ[i*4+3]&0xf0)|(shortUZ[j*4+3]&0xf);
                        Lrotl(&oneElement,11+24-32);
                        longUZ[768+i*16+j] = oneElement;
                }
        }
	SZZS(oneElement);
}




void* gostc_sym_MacOpen()
{
	gostc_sym_mac_t* str = gostc_malloc(sizeof(gostc_sym_mac_t));
	if(!str)
		return NULL;
/*	bzero(str, sizeof(gostc_sym_mac_t)); - already zeroed be gostc_malloc */
	return str;
}

void gostc_sym_MacReset(void* ctx)
{
	bzero(ctx, sizeof(gostc_sym_mac_t));
	return;
}



#define MACBLOCK 8
int gostc_sym_DoMac(void* par, void* info, uint32_t size, uint32_t* aKey,
	uint32_t* aMask, uint32_t* CurUZ)
{
	gostc_sym_mac_t* p = (gostc_sym_mac_t*)par;
	uint32_t *pInfo;
	uint32_t blocks;
	uint32_t i = 0;
	uint32_t tail;
	int rc = 1;

	if(!p || !aKey || !aMask || !CurUZ) return 1;
	if (!size) return 0;
	if (!info) return 1;

        p->total+=size;

	/* check if size < MACBLOCK - left */
	if(size < MACBLOCK - p->left) {
		memcpy(p->temp + p->left, info, size);
		p->left = p->left + size;
		rc = 0;
		goto end;
	}

	memcpy(p->temp+p->left,info, MACBLOCK - p->left);

	pInfo = (uint32_t*)p->temp;
	p->B ^= (*pInfo++);
	p->A ^= (*pInfo++);
	gostc_sym_C16(&p->A,&p->B,aKey,aMask,CurUZ);


	pInfo = (uint32_t*)((uint8_t*)info + MACBLOCK - p->left);
	blocks = (size - MACBLOCK + p->left)/MACBLOCK;
	tail = (size - MACBLOCK + p->left)%MACBLOCK;

	for(i = 0;i<blocks;i++) {
		p->B^=(*pInfo++);
		p->A^=(*pInfo++);
		gostc_sym_C16(&p->A,&p->B,aKey,aMask,CurUZ);
	}

	memset(p->temp, 0, MACBLOCK);
	p->left = 0;
	if(tail)
	{
		memcpy(p->temp,pInfo,tail);
		p->left = tail;
	}
	rc = 0;
end:
	SZZS(blocks);
	SZZS(tail);
	return rc;
}

int gostc_sym_MacClose_plain(void* par, void *mac, uint32_t* aKey, uint32_t* aMask, 
	uint32_t* CurUZ)
{
	gostc_sym_mac_t* p = (gostc_sym_mac_t*)par;
	uint32_t *pInfo;
/*	uint32_t b[16];*/
	int rc = 1;

	if(!p) return 1;
	if (!mac || !aKey || !aMask || !CurUZ) {
		rc = 1;
		goto end;
	}
/*	ZZ(b); */
        if(p->total<16) {
/*		if(p->left) memcpy(b,p->temp,p->left); */
		pInfo = (uint32_t*)p->temp;
		p->B ^= (*pInfo++);
		p->A ^= (*pInfo++);
		gostc_sym_C16(&p->A,&p->B,aKey,aMask,CurUZ);
		p->B ^= (*pInfo++);
		p->A ^= (*pInfo++);
		gostc_sym_C16(&p->A,&p->B,aKey,aMask,CurUZ);
	} else {
		if(p->left) {
			pInfo = (uint32_t*)p->temp;
			p->B^=(*pInfo++);
			p->A^=(*pInfo++);
			gostc_sym_C16(&p->A,&p->B,aKey,aMask,CurUZ);
		}
        }
	*(uint32_t*)mac = p->B;
	rc = 0;
end:
/*	ZZ(b); */
	return rc;
}

int gostc_sym_MacClose(void* par, void *mac, uint32_t* aKey, uint32_t* aMask, 
	uint32_t* CurUZ)
{
	int rc = 1;
	
	if(!par) return 1;
	if (!mac || !aKey || !aMask || !CurUZ) {
		rc = 1;
		goto end;
	}
	rc = gostc_sym_MacClose_plain(par, mac, aKey, aMask, CurUZ);
end:
/*	bzero(p, sizeof(gostc_sym_mac_t)); - zeroed by gostc_free */
	gostc_free(par);
	return rc;
}

/* /////////////////////////////// */
#if 0
int _tmain(int argc, _TCHAR* argv[])
{
	
	uint32_t key[8] = {0x31345521,0xffa3576a,0x1,0x1,0,0,0,0x22222222};
uint32_t data[10] = {11,12,13,4,5,6,7,8,9,10};

uint32_t InnerUZ[] =
{
 0x25AEAAFD,
 0x825489C5,
 0x9E332527,
 0xD698B79A,
 0xF38971EE,
 0x4CFF963F,
 0x6D0BDCA3,
 0x7AED53DB,
 0x59154888,
 0x07DC6EB1,
 0x3B420069,
 0xAF661446,
 0xC1C73F04,
 0x18BAEB52,
 0xB021FD1C,
 0xE470C270
};

uint32_t LongUZ[1024] = {0};
	
	
	uint32_t mac = 0,mac2 = 0;
	setUZ((uint8_t*)InnerUZ,LongUZ);
	LPVOID s = MacOpen();

	uint32_t amask[8] = {2,1,2,1,2,2,2,2};
	
	DoMac(s,data,4,key,amask,LongUZ);
	DoMac(s,data+1,36,key,amask,LongUZ);

	MacClose(s,&mac2,key,amask,LongUZ);

	if(!s)
		printf("Error opening struct in mac\n");
	else
		printf("Opened\n");


	return 0;
}
#endif

