#include "gostc_asym.h"
#include "gostc_asym_vko.h"
#include "gostc_hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <sys/time.h>

/* Uncomment for random testing*/


typedef unsigned char u_int8_t;
#define DBG1 printf


/* First preset seeds*/
#define SEED1 0x55555555
#define SEED2 0xCCCCCCCC
#define CPROSET	1

const uint32_t kTest[] = 
{0xCE5BBB62,0x3475122F,0x94128B9F,0x4FC6AFC2,
0xA61EE031,0x3FA4D560,0x43B22A6F,0x769C1995};

const uint32_t hTest[] = {0xBEE060DB,0x43E7CC88,0x24F23DD7,0xD1E7FAD2,
	0x92AEF5E4,0x224141CB,0x8690AC02,0x54B6DD0A
};


const uint32_t kTest16[16] =
	{0x1ae679f1,0xa3af71bb,0xd14cf70e,0x212273a6,0x11842286,0x44340060,0xa44b3e79,0x86748ed7,
	0x364f3658,0xd455986e,0x0b39d019,0x94631212,0xc6801496,0xcc570456,0xb1410fea,0x0359e7f4};

const uint32_t hTest16[16] =
	{0x72075b8c,0xc6777d29,0x1d560c4f,0x407adedb,0x647c5d5a,0x4339976c,0x6593f441,0x7184ee53,
	0x35b0c591,0xa71d1470,0xc170c533,0x1b09b6f9,0x4d8dab53, 0x5c4f4a7c,0xacc9e061,0x3754f3cf
};

/***************************************************************************/
void dump(void* ptr,int len)
{
	int i, j;
	unsigned char c, *buf;
	unsigned int cc;

	j = 0;
	buf = ptr;
	for (i = 0; i < len; i++) {
		if (j > 15) {
			j = 0;
			printf("\n");
		}
		c = buf[i];
		cc = c;
		printf("%.2X ", cc);
		j++;
	}
	printf("\n");
}
/***************************************************************************/
#define GOSTC_ASYM_TEST_SPEED
/**************************************************************************/

int gostc_asym_TestLibSpeed(double *speed, double *speed2, int n)
{
	gostc_CR_SYSTEM64 *params;
	uint32_t pkTest[32];
	uint32_t sk[16],sgn16[32],rnd[16];
	uint32_t fix1,fix2;
	struct timeval t,t2;
	int i;
		
	bzero(sk,sizeof(sk));
	bzero(sgn16,sizeof(sk));
	bzero(pkTest,sizeof(sk));
	
	if((n<5)||(n>7))
		return 0;

	params = gostc_asym_LoadParams(n);
	if(!params)
	{
		return 0;
	}
	
	srand(0);
rept1:
	for(i=0;i<16;i++)
		rnd[i] = (rand()<<16) + rand();
	if(gostc_asym_GenSecretKey(params, (uint8_t*)rnd, sk)!= 1)
	{
		goto rept1;
	}
	
	gostc_asym_GenPublicKey(params, sk, pkTest);
	
	/*Сгенерили к нему открытую пару*/
	if (gettimeofday(&t, NULL)) {printf("Error gettimeofday 1\n");}
	/*считаем скорость вычисления ЭЦП */
	for(i=0;i<500;i++)
		gostc_asym_MakeSign(params,(uint8_t*)kTest16,(void*)&hTest16,sk,sgn16);
		
	
	
	if (gettimeofday(&t2, NULL)) {printf("Error gettimeofday 2\n");}
	fix1 = (uint32_t)(t2.tv_sec - t.tv_sec);
	if(t.tv_usec <= t2.tv_usec)
		fix2 = (uint32_t)(t2.tv_usec - t.tv_usec)%1000000;
	else
	{
		fix2 = (uint32_t)(1000000 - t.tv_usec + t2.tv_usec); 
		if(fix1)
			fix1--;
	}
	printf("Make sign*500: fix1=%u, fix2=%u\n", fix1, fix2);
	*speed = (double)(500) / (double)((double)fix1 + (double)(fix2)/1000000);
	
	if(gostc_asym_CheckSign(params,(uint8_t*)pkTest,(uint8_t*)hTest16,(uint8_t*)sgn16)!=1)
	{
		printf("check sign - INCORRECT!!!\n");
		gostc_asym_FreeParams(params);
		return 0;
	}

	/*считаем скорость проверки ЭЦП */
	if (gettimeofday(&t, NULL)) {printf("Error gettimeofday 1\n");}
	
	for(i=0;i<500;i++)
		gostc_asym_CheckSign(params,(uint8_t*)pkTest,(uint8_t*)hTest16,(uint8_t*)sgn16);
	if (gettimeofday(&t2, NULL)) {printf("Error gettimeofday 2\n");}
	
	fix1 = (uint32_t)(t2.tv_sec - t.tv_sec);
	if(t.tv_usec <= t2.tv_usec)
		fix2 = (uint32_t)(t2.tv_usec - t.tv_usec)%1000000;
	else
	{
		fix2 = (uint32_t)(1000000 - t.tv_usec + t2.tv_usec); 
		if(fix1)
			fix1--;
	} 
	printf("Check sign*500: fix1=%u, fix2=%u\n", fix1, fix2);
	*speed2 = (double)(500) / (double)((double)fix1 + (double)(fix2)/1000000);
	
	gostc_asym_FreeParams(params);
	return 1;
}

/*******************************************************************/

int gostc_asym_TestLibSpeed256(double *speed, double *speed2, int n)
{
	gostc_CR_SYSTEM *params;
	uint32_t pkTest[16];
	uint32_t sk[8],sgn16[16],rnd[8];
	uint32_t fix1,fix2;
	struct timeval t,t2;
	int i;
		
	bzero(sk,sizeof(sk));
	bzero(sgn16,sizeof(sk));
	bzero(pkTest,sizeof(sk));
	
	if((n<1)||(n>4))
		return 0;

	params = gostc_asym_LoadParams(n);
	if(!params)
	{
		return 0;
	}
	
	srand(0);
rept1:
	for(i=0;i<8;i++)
		rnd[i] = (rand()<<16) + rand();
	if(gostc_asym_GenSecretKey(params, (uint8_t*)rnd, sk)!= 1)
	{
		goto rept1;
	}
	
	gostc_asym_GenPublicKey(params, sk, pkTest);
	
	/*Сгенерили к нему открытую пару*/
	if (gettimeofday(&t, NULL)) {printf("Error gettimeofday 1\n");}
	/*считаем скорость вычисления ЭЦП */
	for(i=0;i<500;i++)
		gostc_asym_MakeSign(params,(uint8_t*)kTest,(void*)&hTest,sk,sgn16);
		
	
	
	if (gettimeofday(&t2, NULL)) {printf("Error gettimeofday 2\n");}
	fix1 = (uint32_t)(t2.tv_sec - t.tv_sec);
	if(t.tv_usec <= t2.tv_usec)
		fix2 = (uint32_t)(t2.tv_usec - t.tv_usec)%1000000;
	else
	{
		fix2 = (uint32_t)(1000000 - t.tv_usec + t2.tv_usec); 
		if(fix1)
			fix1--;
	}
		
	printf("Make sign*500: fix1=%u, fix2=%u\n", fix1, fix2);
	*speed = (double)(500) / (double)((double)fix1 + (double)(fix2)/1000000);
	
	if(gostc_asym_CheckSign(params,(uint8_t*)pkTest,(uint8_t*)hTest,(uint8_t*)sgn16)!=1)
	{
		printf("check sign - INCORRECT!!!\n");
		gostc_asym_FreeParams(params);
		return 0;
	}

	/*считаем скорость проверки ЭЦП */
	if (gettimeofday(&t, NULL)) {printf("Error gettimeofday 1\n");}
	
	for(i=0;i<500;i++)
		gostc_asym_CheckSign(params,(uint8_t*)pkTest,(uint8_t*)hTest,(uint8_t*)sgn16);
	if (gettimeofday(&t2, NULL)) {printf("Error gettimeofday 2\n");}
	
	fix1 = (uint32_t)(t2.tv_sec - t.tv_sec);
	if(t.tv_usec <= t2.tv_usec)
		fix2 = (uint32_t)(t2.tv_usec - t.tv_usec)%1000000;
	else
	{
		fix2 = (uint32_t)(1000000 - t.tv_usec + t2.tv_usec); 
		if(fix1)
			fix1--;
	}
	printf("Check sign*500: fix1=%u, fix2=%u\n", fix1, fix2);
	*speed2 = (double)(500) / (double)((double)fix1 + (double)(fix2)/1000000);
	
	gostc_asym_FreeParams(params);
	return 1;
}



/**************************************************************************/
int main(int argc, char ** argv)
{

#ifdef GOSTC_ASYM_TEST_SPEED
	double sp,sp2;
#endif
	
	if (!gostc_asym_TestLib()) {
		printf("Self-test of GOST R 3410-2001 library - FAILED.\n");
		return 1;
	}
	else
	{
		printf("Self-test of GOST R 3410-2001 library - CORRECT!\n");
	}
#ifdef GOSTC_ASYM_TEST_SPEED
	printf("Speed-test gost2012-256-testSet\n");
	if(!gostc_asym_TestLibSpeed256(&sp, &sp2,1))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}	

	printf("Speed-test gost2012-256-CproASet\n");
	if(!gostc_asym_TestLibSpeed256(&sp, &sp2,2))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
	printf("Speed-test gost2012-256-CproBSet\n");
	if(!gostc_asym_TestLibSpeed256(&sp, &sp2,3))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
	
	printf("Speed-test gost2012-256-CproCSet\n");
	if(!gostc_asym_TestLibSpeed256(&sp, &sp2,4))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
	printf("Speed-test gost2012-512-testSet\n");
	if(!gostc_asym_TestLibSpeed(&sp, &sp2,5))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
		return 0;
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
	printf("Speed-test gost2012-512-ParamSetA\n");
	if(!gostc_asym_TestLibSpeed(&sp, &sp2,6))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
		return 0;
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
	printf("Speed-test gost2012-512-ParamSetB\n");
	if(!gostc_asym_TestLibSpeed(&sp, &sp2,7))
	{
		printf("Speed-test of GOST R 3410-2012 library failed!\n");
		return 0;
	}
	else
	{
		printf("Speed-test is correct. Frequency making: %10.2f Hz; checking: %10.2f Hz\n",sp, sp2);
	}
#endif
	
	return 0;
}
/***************************************************************************/

