#ifndef _PKCS15CONTAINER_H_INCLUDED_
#define _PKCS15CONTAINER_H_INCLUDED_

/* Library Version: */
#define P15_2001_FTS_MAJOR_VERSION	3
#define P15_2001_FTS_MINOR_VERSION	2

#define P15_2012_FTS_MAJOR_VERSION	4
#define P15_2012_FTS_MINOR_VERSION	4
/* For more info about versions see pkcs15.asn file */


#include "gostc_common.h"

/* IMPORTANT! For allocating and freeing p15_ objects - use only functions below.
 * Do not allocate or free objects manually.
 */

#include "p15_mem.h" /* Allocation and freeing functions with zeroing. */


/* Container formats */
#define P15_FORMAT_ANY		0
#define P15_FORMAT_2001		1	/* P15-2001 format */
#define P15_FORMAT_2012		2	/* P15-2012 format */


/* Error types (for p15_info_t struct) */
#define P15_OK			0
#define P15_WARNING		1
#define P15_ERROR		2
#define P15_INTERNAL_ERROR	3

/* Error type policy for p15_save_token/p15_load_token:
 * P15_OK - not used in p15_info_t.type
 * P15_WARNING - Input data is ok, but some structures were skipped, because not supported
 * P15_ERROR - Input data has errors (erroneous structure in token, invalid password, etc.)
 * P15_INTERNAL_ERROR - error in p15_ library.
 */

/* Return codes */

#define P15_RC_OK				0
/* No errors, no warnings */

#define P15_RC_WARNINGS				1
/* No errors, warnings present */

#define P15_RC_INTERNAL_ERROR			3

#define P15_RC_WRONG_PASSWORD			10
#define P15_RC_INVALID_ARGS			11
#define P15_RC_CERT_DECODING_FAILED		12
#define P15_RC_INVALID_ISSUER_RDN		13
#define P15_RC_CERT_NOT_FOUND			14
#define P15_RC_CERT_ID_IN_USE			15
#define P15_RC_INVALID_SUBJECT_RDN		16
#define P15_RC_PRIVATE_KEY_NOT_FOUND		17
#define P15_RC_PRIVATE_KEY_ID_IN_USE		18
#define P15_RC_CRL_ID_IN_USE			19
#define P15_RC_CRL_DECODING_FAILED		20
#define P15_RC_CRL_NOT_FOUND			21
#define P15_RC_RNG_NOT_INITIALIZED		22
#define P15_RC_LABEL_TOO_LONG			23
#define P15_RC_ASN1C_ERROR			24
#define P15_RC_INVALID_OID_STR			25
#define P15_RC_PKCS5_FAILED			26
#define P15_RC_CRYPTOLIB_IMIT			27
#define P15_RC_BUFFER_TOO_SMALL			28
#define P15_RC_DER_DECODING_FAILED		29
#define P15_RC_DER_NOT_COMPLETE			30
#define P15_RC_INVALID_TOKEN			31
#define P15_RC_IGNORED_STRUCTURE		32
#define P15_RC_URANDOM_ERROR			33
#define P15_RC_INVALID_PBKDF2_PARAMS		34
#define P15_RC_NO_PWINFO			35
#define P15_RC_NO_AUTH_AND_DECRYPT		36
#define P15_RC_UNSUPPORTED_ALGORITHM		37
#define P15_RC_NO_RANDOM_CONTAINER		38
#define P15_RC_INV_OBJ_FLAGS			39
#define P15_RC_INV_FACTOR_TS_VERSION		40
#define P15_RC_NO_SUBJECT_FOR_PR_KEY		41
#define P15_RC_NO_PRIVATE_KEY_DATA		42
#define P15_RC_INVALID_OBJECT			43
#define P15_RC_KEY_NOT_FOUND			44
#define P15_RC_PUBLIC_KEY_ID_IN_USE		45
#define P15_RC_PUBLIC_KEY_NOT_FOUND		46
#define P15_RC_NO_SUBJECT_FOR_PUB_KEY		47
#define P15_RC_NO_OBJECT_DATA			48
#define P15_RC_CRYPTOLIB_MASK			49
#define P15_RC_INVALID_BIT_STRING		50
#define P15_RC_DUPLICATE_CERT			51
#define P15_RC_DUPLICATE_CRL			52
#define P15_RC_2012_KEYS_EXIST			53
#define P15_RC_2012_KEY_VS_2001_FORMAT		54
#define P15_RC_CRYPTOLIB_HASH			55
#define P15_RC_INVALID_CONTAINER_FORMAT		56
#define P15_RC_NO_SECRET_KEY_DATA		57
#define P15_RC_ID_IN_USE			58
#define P15_RC_SECRET_KEY_NOT_FOUND		59
#define P15_RC_89UZ_Z_VS_2001_FORMAT		60
/* When adding a new error, don't forget to modify p15_rc_dsc content */
#define P15_RC_MAX				60

/* Array of text descriptions of return codes */
#ifdef __cplusplus
extern "C" {
#endif

extern char p15_rc_dsc[P15_RC_MAX + 1][500];

#ifdef __cplusplus
}
#endif



#define P15_LABEL_MAX		255
#define P15_EXT_ID_MAX		255

#define P15_KEY_MASKS_MAX	50

#define P15_MORE_RANDOM_MAX	90000

#define P15_PAGE		65536


/* Object types */
#define P15_OT_UNKNOWN		0
#define P15_OT_PRIVKEY		1
#define	P15_OT_PUBKEY		2
#define P15_OT_CERT		3
#define P15_OT_TRUSTED_CERT	4
#define P15_OT_CRL		5
#define P15_OT_OIDDO		6
#define P15_OT_SECKEY		7


/* Key usage flags */
#define P15_KU_ENCRYPT		0x8000
#define P15_KU_DECRYPT		0x4000
#define P15_KU_SIGN		0x2000
#define P15_KU_SIGN_RECOVER	0x1000
#define P15_KU_WRAP		0x0800
#define P15_KU_UNWRAP		0x0400
#define P15_KU_VERIFY		0x0200
#define P15_KU_VERIFY_RECOVER	0x0100
#define P15_KU_DERIVE		0x0080
#define P15_KU_NON_REPUDIATION	0x0040

/* Key access flags */
#define P15_KA_ABSENT		0x00 /* No KeyAccessFlags in container */
#define P15_KA_PRESENT		0x01 /* KeyAccessFlags exist, all flags cleared */
#define P15_KA_SENSITIVE	0x80 /* Plain key should not be displayed */
#define P15_KA_EXTRACTABLE	0x40 /* Key is exportable */
#define P15_KA_ALWAYSSENSITIVE	0x20 /* Key has always been sensitive */
#define P15_KA_NEVEREXTRACTABLE	0x10 /* Key has never been exportable */
#define P15_KA_LOCAL		0x08 /* Key locally generated by this smart card 
					(not applicable for software token) */
#define P15_KA_DEFAULT		(P15_KA_SENSITIVE | P15_KA_EXTRACTABLE | \
				 P15_KA_ALWAYSSENSITIVE)
#define P15_KA_MASK		0xF9

/* PKI paramsets */
#define P15_PKIPAR_CPRO_A		1
#define P15_PKIPAR_CPRO_B		2
#define P15_PKIPAR_CPRO_C		3
#define P15_PKIPAR_CPRO_XCHA		4
#define P15_PKIPAR_CPRO_XCHB		5
/* For P15-2012 only */
#define P15_PKIPAR_TC26_512_TEST	6
#define P15_PKIPAR_TC26_512_A		7
#define P15_PKIPAR_TC26_512_B		8

/* Encr S-boxes */
#define	P15_89UZ_CPRO_A		1
#define	P15_89UZ_CPRO_B		2
#define	P15_89UZ_CPRO_C		3
#define	P15_89UZ_CPRO_D		4
#define P15_89UZ_CPRO_Z		5 /* Only for P15-2012 */


/* OIDs */
#define P15_OID_RNG		"1.3.6.1.4.1.13312.503.1.1"
#define P15_OID_CRL		"1.3.6.1.4.1.13312.503.1.2"
#define P15_OID_FACTOR_VER	"1.3.6.1.4.1.13312.503.1.3"
#define P15_OID_ABSTRACT_DATA	"1.3.6.1.4.1.13312.503.1.4"
#define P15_OID_AA_ID_MDIGEST	"1.2.840.113549.1.9.4"
#define P15_OID_AA_ID_CTYPE	"1.2.840.113549.1.9.3"
#define P15_OID_AA_CTYPE	"1.2.840.113549.1.15.3.1"
#define P15_OID_ENC_CTYPE	"1.2.840.113549.1.7.1"
#define P15_OID_PBKDF2		"1.2.840.113549.1.5.12"
#define P15_OID_CONTENT_TYPE	"1.2.840.113549.1.15.3.1"
#define P15_OID_GOST_R3411_94	"1.2.643.2.2.9"
#define P15_OID_GOST_R3411_2012_512	"1.2.643.7.1.1.2.3"
#define P15_OID_HMAC_GOST_94	"1.2.643.2.2.10"
#define P15_OID_HMAC_GOST_2012	"1.2.643.7.1.1.4.2"
#define P15_OID_GOST_89		"1.2.643.2.2.21"
#define P15_OID_GOST_KEY_WRAP	"1.2.643.2.2.13.1"
#define P15_OID_GOST_ENCR_IMIT	"1.2.643.2.4.3.2.2"
#define P15_OID_ENCR_CPRO_A	"1.2.643.2.2.31.1"
#define P15_OID_ENCR_CPRO_B	"1.2.643.2.2.31.2"
#define P15_OID_ENCR_CPRO_C	"1.2.643.2.2.31.3"
#define P15_OID_ENCR_CPRO_D	"1.2.643.2.2.31.4"
#define P15_OID_ENCR_CPRO_Z	"1.2.643.7.1.2.5.1.1"
#define P15_OID_HASH_CPRO	"1.2.643.2.2.30.1"
#define P15_OID_PKI_CPRO_A	"1.2.643.2.2.35.1"
#define P15_OID_PKI_CPRO_B	"1.2.643.2.2.35.2"
#define P15_OID_PKI_CPRO_C	"1.2.643.2.2.35.3"
#define P15_OID_PKI_CPRO_XCHA	"1.2.643.2.2.36.0"
#define P15_OID_PKI_CPRO_XCHB	"1.2.643.2.2.36.1"
#define P15_OID_PKI_TC26_TEST	"1.2.643.7.1.2.1.2.0"
#define P15_OID_PKI_TC26_A	"1.2.643.7.1.2.1.2.1"
#define P15_OID_PKI_TC26_B	"1.2.643.7.1.2.1.2.2"


/* PBKDF2 */
#define P15_P5_DEFAULT_SALT_LEN	32
#define P15_P5_DEFAULT_C	2000



typedef struct p15_id_t p15_id_t;
struct p15_id_t {
	p15_id_t*	next; /* for id lists (for find operations) */
	uint32_t	id; /* used when ID is 4 octets (big-endian in OCTET STRING) */
	int		ext_id_len; /* 0 if id is used, !0 if ext_id is used */
	uint8_t		ext_id[P15_EXT_ID_MAX];
};


typedef struct p15_utctime_t {
	uint16_t	year;
	uint8_t		month;
	uint8_t		day;
	uint8_t		hours;
	uint8_t		minutes;
	uint8_t		seconds;
} p15_utctime_t;


typedef struct p15_der_t {
	uint8_t*	ptr;
	size_t		len;
} p15_der_t;


typedef struct p15_oid_t {
	uint32_t*	arcs;	/* array of oid parts */
	int		arcnum;	/* number of oid parts */
} p15_oid_t;

typedef struct p15_certificate_t p15_certificate_t;
struct p15_certificate_t {
	char			label[P15_LABEL_MAX + 1];
	p15_id_t		id;
	uint8_t*		ptr;
	size_t			len;
	int			ca;
	p15_der_t*		subject;
	p15_der_t*		issuer;
	p15_der_t*		serial_number;
};

typedef struct p15_private_key_t p15_private_key_t;
struct p15_private_key_t {
	char			label[P15_LABEL_MAX + 1];
	p15_id_t		id;

	uint8_t**		key_masks;	
	/* array of pointers to key masks - may be NULL after p15_load_token 
	 * if wrong password. Fq field multiplication masking.
	 * Key mask may be 256bit or 512bit depending on key_parameters.
	 */

	size_t			mask_count;

	int			to_remask;
	/* If !0 - the key will be remasked during p15_save_token.
	 * If the key is newly created - 0 by default
	 * p15_load_token sets the flag to 1.
	 * See p15_remask_setup, p15_remask_private_key functions.
	 */

	uint16_t		key_usage;	/* P15_KU_ ORed constants */

	uint8_t			key_access_flags;
	/* may be P15_KA_ABSENT, P15_KA_PRESENT, other P15_KA_ ORed constants. 
	 * See P15_KA_ constants above.
	 */

	int			key_parameters;	/* P15_PKIPAR_ const */
	p15_utctime_t*		start_date;	/* optional */
	p15_utctime_t*		end_date;	/* optional */
	p15_der_t*		subject;	/* optional */

};

typedef struct p15_secret_key_t p15_secret_key_t;
struct p15_secret_key_t {
	char			label[P15_LABEL_MAX + 1];
	p15_id_t		id; /* Should not intersect with other object types */

	uint8_t**		key_masks;	
	/* array of pointers to key masks - may be NULL after p15_load_token 
	 * if wrong password. XOR-masking.
	 * Key/mask is 256bit.
	 */

	size_t			mask_count;

	int			to_remask;
	/* If !0 - the key will be remasked during p15_save_token.
	 * If the key is newly created - 0 by default
	 * p15_load_token sets the flag to 1.
	 * See p15_remask_setup, p15_remask_secret_key functions.
	 */

	uint16_t		key_usage;	/* P15_KU_ ORed constants */

	uint8_t			key_access_flags;
	/* may be P15_KA_ABSENT, P15_KA_PRESENT, other P15_KA_ ORed constants. 
	 * See P15_KA_ constants above.
	 */

	int			key_parameters;	/* P15_89UZ_ const */
	p15_utctime_t*		start_date;	/* optional */
	p15_utctime_t*		end_date;	/* optional */

};

typedef struct p15_public_key_t p15_public_key_t;
struct p15_public_key_t {
	char			label[P15_LABEL_MAX + 1];
	p15_id_t		id;
	uint8_t			key[128];
	uint16_t		key_usage;	/* P15_KU_ ORed constants */
	int			key_parameters; /* P15_PKIPAR_ const */
	p15_utctime_t*		start_date;	/* optional */
	p15_utctime_t*		end_date;	/* optional */
	p15_der_t*		subject;	/* optional */
};
	
typedef struct p15_crl_t p15_crl_t;
struct p15_crl_t {
	char		label[P15_LABEL_MAX + 1];
	p15_id_t	id;
	uint8_t*	ptr;
	size_t		len;
	p15_der_t*	issuer;
	p15_der_t*	this_update;
	p15_der_t*	next_update; /* optional */
};

typedef struct p15_oiddo_t p15_oiddo_t;
struct p15_oiddo_t {
	char		label[P15_LABEL_MAX + 1];
	p15_oid_t*	oid;
	uint8_t*	ptr; /* May be NULL for encrypted objects        */
	size_t		len; /* after p15_load_token with wrong password */
	int		to_encrypt;
};

typedef struct p15_object_t p15_object_t;
struct p15_object_t {
	p15_object_t*	prev;
	p15_object_t*	next;
	int 		type; /* P15_OT_ const */
	void*		obj;  
/* may be:
 *  - p15_certificate_t*
 *  - p15_private_key_t*
 *  - p15_public_key_t*
 *  - p15_secret_key_t*
 *  - p15_crl_t*
 *  - p15_oiddo_t*
 *  ...
 */

};

/* /////////////////////////////// MAIN CONTEXT //////////////////////////// */

typedef struct p15_ctx_t {

/* IMPORTANT. The ctx is read-only. Do not write anything manually to ctx.
 *            Use functions described below.
 */
 	/* Container format */
	int		p15_format; /* P15_FORMAT_* const. */

	/* See p15_set_format, p15_load_token, p15_save_token */
	int		version_major, version_minor; /* filled by p15_load_token */

 	/* RNG context */
	int 		rnd_init; /* 0-not init, 1-inited */
	uint8_t 	rnd_iv[8];
	uint8_t*	rnd_val;
	uint8_t*	rnd_msk;
	uint8_t*	rnd_extra;
	int		rnd_extra_len; /* moreRandom read bytes */
	int		rnd_new_extra_len; /* moreRandom written bytes */
	uint8_t*	rnd_new_fix_val; /* rndInit + moreRandom - to write fixed value */

	/* password related */
	char		password_hint[P15_LABEL_MAX + 1];
	int		p5_c; /* iteration count for PBKDF2 */
	int		p5_salt_len;

	/* UZ */
	int		enc_uz;  
	/* P15_89UZ_ const. Default - CPROSET-A. 
	 * p15_load_token overrides the option by the value of:
AuthenticatedData.recipientInfos[0].kekri.keyEncryptionAlgorithm.parameters.encryptionParamSet
	 */

	/* global object list */
	p15_object_t*	objects;

	/* Other internal vars */
	uint32_t 	free_id; /* unallocated ID counter */
	int		fts_ver_detected;

} p15_ctx_t;

/* ////////////////////////////////////////////////////////////////////////// */


typedef struct p15_info_t p15_info_t;
struct p15_info_t {
	p15_info_t*	next;
	int 		type; /* P15_ERROR, P15_WARNING, P15_INTERNAL_ERROR */
	int		code; /* P15_RC_ const */
	char*		text; /* some detailed description */
/* For debug: */
	size_t		consumed; /* bytes consumed by decoder */
	char*		asn_type_name; /* name of type when encode error occurred */
};




#ifdef __cplusplus
extern "C" {
#endif

/* ////////////////////////////// Interface functions ///////////////////////// */

/* Context functions */

extern p15_ctx_t* p15_new_ctx(void);
/* Creates empty context. Random generator not initialized. NULL if error */

extern void p15_free_ctx(p15_ctx_t**);
/* Securely frees context */


/* Container format */
extern int p15_set_format(p15_ctx_t* ctx, int format);
/* Sets (forces) the format of container. See P15_FORMAT_* consts. 
 * When a new ctx created, the ctx->p15_format var is set to P15_FORMAT_ANY.

 * p15_load_token behavior, depending on format set:
 * P15_FORMAT_ANY  - autodetect container format when loading. 
 *                   ctx->p15_format is set to detected format.
 * P15_FORMAT_2001 - read old-format containers only. Error if P15-2012 format.
 * P15_FORMAT_2012 - read new-format containers only. Error if P15-2001 format.

 * p15_save_token behavior, depending on format set:
 * P15_FORMAT_ANY  - set format to P15_FORMAT_2012 and save container in new format.
 * P15_FORMAT_2001 - save container in old format.
 * P15_FORMAT_2012 - save container in new format.

 * p15_add_private_key(2), p15_add_public_key behavior, if adding a key with 512-bit paramset:
 * P15_FORMAT_ANY  - force ctx->p15_format to P15_FORMAT_2012
 * P15_FORMAT_2001 - return error
 * P15_FORMAT_2012 - standard behavior.

 * p15_add_secret_key(2) behavior:
 * P15_FORMAT_ANY  - force ctx->p15_format to P15_FORMAT_2012
 * P15_FORMAT_2001 - return error
 * P15_FORMAT_2012 - standard behavior.

 * p15_set_default_enc_uz behavior, if setting P15_89UZ_CPRO_Z paramset:
 * P15_FORMAT_ANY  - force ctx->p15_format to P15_FORMAT_2012
 * P15_FORMAT_2001 - return error
 * P15_FORMAT_2012 - standard behavior.

 * Returns: 0-ok !0-P15_RC_ const for error
 * An error is returned, if an attempt is performed to set P15_FORMAT_ANY or _2001 when
 * GOST R 34.10-2012 keys exist in ctx.
 */



/* Load Token function */

extern int p15_load_token(p15_ctx_t* ctx, uint8_t* token, size_t token_len, 
	uint8_t* password, int passlen, p15_info_t** errors);
/* Clears old data in ctx. Loads a token. 
 * If random generator not initialized - random value is taken from the container
 * (Before random value is read - key masks are taken from /dev/urandom).
 * The function tries to load as much data as possible. Unencrypted data is always loaded.
 * For encrypted objects - if authentication/decryption fails - loads only 
 * info (labels, IDs, RDNs).
 * The order of objects is preserved as in p15_ctx_t.objects list.
 * Returns: 0 - on success, P15_RC_WARNINGS - if no errors, but warnings exist,
 * P15_RC_ constant for most significant error.
 * Most important errors:
 * P15_RC_WRONG_PASSWORD - Authentication/decryption failed because of incorrect password
 * P15_RC_NO_AUTH_AND_DECRYPT - Authentication/decryption failed because of other reasons:
 * incorrect or incompatible token structure, intentionally corrupted token, etc.
 * P15_RC_INVALID_TOKEN - Authentication succeeded, but token contains erroneous structures.
 * If errors!=NULL - error list is allocated (or appended) with detailed errors
 * (*errors must be freed by p15_free_infos after the call).
 * See also p15_set_format notes above.
 */


/* Common object functions */

extern int p15_rm_object(p15_ctx_t* ctx, p15_object_t* obj);
/* Removes the object from the list and frees it.
 * Ret: 0-ok, !0-P15_RC_ const
 */

/* See also p15_get_next_* functions below */


/* Helper functions */

extern void p15_free_infos(p15_info_t** info_list);
/* Frees list of info messages. */

/* - not needed
extern int p15_set_skeys_mask_count(p15_ctx_t* ctx, int count);
 * Sets mask count for container encryption keys and random init value in the token.
 * count=1 - the key only, 2-key w/mask (default), 3-key w/2masks, ...
 * Ret: 0-ok, !0-P15_RC_ const
 */

extern int p15_set_default_enc_uz(p15_ctx_t* ctx, int uz);
/* Sets default UZ for encrypting contents. uz must be P15_89UZ_ const.
 * Ret: 0-ok, !0-P15_RC_ const
 * Note: p15_load_token overrides the option by the value of:
AuthenticatedData.recipientInfos[0].kekri.keyEncryptionAlgorithm.parameters.encryptionParamSetA
 * This is done to preserve enc_uz setting between p15_load_token/p15_save_token pair.
 */

extern int p15_set_password_hint(p15_ctx_t* ctx, char* pwhint);
/* Sets password hint. strlen(pwhint) must be < 256.
 * If pwhint==NULL - old pwhint is cleared.
 * Ret: 0-ok, !0-P15_RC_ const
 */

extern char* p15_get_password_hint(p15_ctx_t* ctx);
/* Returns password hint or NULL if no hint or error
 * Return value exist within ctx and it should not be freed manually.
 */

extern int p15_set_pbkdf2_params(p15_ctx_t* ctx, uint16_t c, uint16_t saltlen);
/* Sets PKCS5 parameters - iteration count and salt length
 * Ret: 0-ok, !0-P15_RC_ const
 */



/* Random generator functions */

extern int p15_init_random(p15_ctx_t* ctx, uint8_t* rnd, uint8_t* msk);
/* Initializes random generator. rnd and msk (mask) should be 32 bytes long. 
 * rnd and msk are copied.
 * Ret: 0-ok, !0-P15_RC_ const
 * The function should be called if p15_load_token function was not called or
 * the token does not contain random container
 */

extern int p15_get_random(p15_ctx_t* ctx, uint8_t* out_buf, int len);
/* Returns len bytes of random data to *out_buf
 * RNG must be initialized
 * Ret: 0-ok, !0-P15_RC_ const
 */

extern int p15_random_setup(p15_ctx_t* ctx, uint8_t* fixed_value, int len);
/* Specifies how RandomInitValue object will be written to the token.
 * Should be called just before p15_save_token.
 * p15_load_token overrides this settings:
 * (equivalent to p15_random_setup(ctx, NULL, 64 + sizeof(RandomInitValue.moreRandom)) )
 * Default (if not called) - RandomInitValue.randomInit will be filled from RNG. 
 *  No moreRandom data written.
 * fixed_value = NULL, len = 64 - equivalent to default.
 * fixed_value = NULL, len > 64 - randomInit filled from RNG. 
 *  (len - 64) bytes is written to moreRandom from RNG.
 * If fixed_value!=NULL - it should be len bytes long. 
 *  In this case fixed_value data will be written to randomInit + moreRandom.
 * len should be: 64 <= len <= P15_MORE_RANDOM_MAX
 * Ret: 0-ok, !0-P15_RC_ const
 */


extern int p15_get_extra_random_len(p15_ctx_t* ctx);
/* Returns number of extra random bytes read from RandomInitValue.moreRandom field
 * Should be called after p15_load_token
 * 0 if failed or if no moreRandom field
 */

extern int p15_get_extra_random(p15_ctx_t* ctx, uint8_t* buf, int* len);
/* Copies len bytes from RandomInitValue.moreRandom to buf.
 * Actual number of bytes copied returned back to len (if moreRandom < len). May be 0
 * Ret: 0-ok (even if *len==0), !0-P15_RC_ const
 */

/* ID functions */

extern p15_id_t* p15_new_id(uint32_t id);
/* Creates the id structure from short id */

extern p15_id_t* p15_new_ext_id(uint8_t* id, int len);
/* Creates long id. len should be < 256 */

extern void p15_free_id(p15_id_t** id);
/* Frees the structure (or a list). */

extern int p15_id_cmp(p15_id_t* id1, p15_id_t* id2);
/* Compares IDs. 
 * Returns 0 if equal. 1 if not equal. P15_RC_ const if error.
 */



/* RDN and DER functions */

extern p15_der_t* p15_new_rdn(uint8_t* rdn_der, int len);
/* Decodes rdn_der to check if RDN valid and allocates new DER object */

extern p15_der_t* p15_new_der(uint8_t* der, int len);
/* Allocates new DER object. der is copied. */

extern void p15_free_der(p15_der_t**);
/* Frees DER object. Only for freeing standalone objects. 
 * DERs in ctx are freed automatically.
 */

extern int p15_der_cmp(p15_der_t* der1, p15_der_t* der2);
/* Compares DERs.
 * Returns 0 if equal. 1 if not equal. P15_RC_ const if error.
 */




/*****************************************************************************

 * Adding functions like p15_add_cert, p15_add_private_key, p15_add_crl, etc.
 * add objects of different types to the global list (p15_ctx_t.objects).
 * The mixed-object-type order is preserved in the token and is restored after 
 * loading.

 *****************************************************************************/


/* Certificate functions */

extern int p15_add_cert(p15_ctx_t* ctx, int trusted, int ca, p15_id_t* id, p15_id_t** newid, 
	char* label, uint8_t* cert, size_t certlen, p15_info_t** errors);
/* adds new certificate to container.
 * trusted - if 1 - certificate is added to trustedCertificates list, 0-common list
 * ca - 0 - end-user (non-CA) certificate, 1 - CA certificate
 * id - if NULL - new id will be created 
 * and returned to newly allocated *newid (if newid!=NULL).
 * If id is specified - given cert stored with given id. 
 * (If old cert with same id exists - error).
 * (id struct is copied, and the source id struct should be freed if not needed).
 * IDs should be equal for pairs: certificate - private keys, CA certificate - CRL.
 * label - (optional) sets text label of cert container (max 255 chars). (label is copied).
 * cert should be in DER. (copied).
 * If errors!=NULL - error list is allocated (or appended) with detailed errors
 * (*errors must be freed by p15_free_infos after the call).
 * Ret: 0-ok, P15_RC_WARNINGS - if no errors, but warnings exist, >1 - P15_RC_ const 
 * of most significant error
 */

extern int p15_rm_cert(p15_ctx_t* ctx, p15_id_t* id);
/* Removes certificate(s) from the list(s) by given id.
 * Returns P15_RC_ const.
 */

extern int p15_find_certs_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found);
/* Searches certs by subject and allocates the 'found' list of IDs.
 * If not found - *found=NULL
 * found list should be freed by p15_free_id.
 * Ret: 0-ok (even if not found), P15_RC_ constant if error
 */

extern p15_certificate_t* p15_get_cert(p15_ctx_t* ctx, p15_id_t* id);
/* Returns certificate structure by id. NULL if not found or error.
 * The returned structure is read-only and should not be freed manually.
 * It will be freed by p15_free_ctx.
 */

extern p15_certificate_t* p15_get_next_cert(p15_ctx_t* ctx, p15_object_t** it, int trusted);
/* Iterates next certificate from the global list. 
 * *it=NULL - to start iteration.
 * Returns NULL if no more certificates or error.
 */

extern p15_certificate_t* p15_copy_cert(p15_certificate_t*);
/* Alocates new independent object and makes a smart copy 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_cert.
 */

extern void p15_free_cert(p15_certificate_t**);
/* Frees an independent certificate object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */




/* Time functions */

extern p15_utctime_t* p15_new_time(int year, int month, int day, 
	int hours, int minutes, int seconds);
/* Creates new time object. Returns NULL if error. Time should be UTC. */

extern void p15_free_time(p15_utctime_t**);
/* Frees time object */



/* Private key functions */

extern int p15_add_private_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors,
	int mask_count, uint8_t* key, ...);
/* Adds private key. 
 * About IDs - see p15_add_certificate
 * All input pointed structures are copied and should be freed manually if not needed.
 * key_usage - P15_KU_ ORed constants
 * key_access_flags - may be P15_KA_ABSENT, P15_KA_PRESENT, other P15_KA_ ORed constants.
 * P15_KA_DEFAULT is recommended. See above about P15_KA_ constants.
 * key_params - P15_PKIPAR_ constant
 * start_date - (optional) - for information about the key lifetime
 * end_date - (optional)
 * mask_count - count of key+masks. 1-key only, 2-key w/mask, 3-key w/2 masks,...
 * key - 32/64 bytes of the (masked) key (little-endian)
 * Masking is based on Fq field multiplication.
 * If mask_count > 1, uint8_t* pointers should follow key arg, pointing to masks.
 * Ret: 0-ok, 1-if only warnings, P15_RC_ const for main error.
 * See also p15_set_format notes above.
 */

extern int p15_add_private_key2(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	int mask_count, uint8_t** key_masks,
	p15_info_t** errors);
/* Same as p15_add_private_key, except key_masks - is an array of key masks:
 * key_mask[0] - masked key
 * key_mask[1] - mask 1
 * key_mask[2] - mask 2 ...
 * All input pointed structures are copied and should be freed manually if not needed.
 */

extern int p15_remask_setup(p15_ctx_t* ctx, p15_private_key_t* prkey,
	p15_secret_key_t* seckey, int to_remask, int mask_count, p15_info_t** errors);
/* Sets/clears to_remask flag of a private key(s). May also do remasking if mask_count>0.
 * See also p15_private_key_t structure.
 * If the flag is set, the key will be remasked during p15_save_token.
 * If !(prkey || seckey) - the flag is modified for all private keys in ctx.
 * If (prkey || seckey) - the flag is modified for the specified key(s).
 * If mask_count>0 - the key(s) is(are) remasked immediately to the new number of masks.
 * If mask_count==1 - the key(s) is(are) unmasked.  
 * Ret: P15_RC_ const.
 */

extern int p15_remask_private_key(p15_ctx_t* ctx, p15_private_key_t* key, int mask_count);
/* Masks/umasks/remasks a private key. 
 * Random generator should be initialized.
 * mask_count should be > 0.
 * If mask_count==1 - the key is unmasked.
 * Ret: P15_RC_ const.
 */ 

extern int p15_set_pkey_subject(p15_ctx_t* ctx, p15_id_t* id, 
	p15_der_t* rdn);
/* Sets subject for a private/public key(s) (for informational uses).
 * The key(s) are found by ID.
 * *rdn is copied.
 * If rdn==NULL - tries to find the certificate with the same ID and gets its subject.
 * Ret: P15_RC_ const
 */

extern int p15_rm_private_key(p15_ctx_t* ctx, p15_id_t* id);
/* Removes private key by id.
 * Ret: P15_RC_ const
 */

extern int p15_find_private_keys_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found);
/* Same as p15_find_certs_by_subject */

extern p15_private_key_t* p15_get_private_key(p15_ctx_t* ctx, p15_id_t* id);
/* Returns read-only structure of private key, found by id. 
 * Or NULL if not found or error. 
 * Returned structure is a part of ctx and should not be freed manually. 
 * If key is not decrypted (wrong password) - key_masks=NULL.
 */

extern p15_private_key_t* p15_get_next_private_key(p15_ctx_t* ctx, p15_object_t** it);
/* Iterates next private key from the global list. 
 * *it=NULL - to start iteration.
 * Returns NULL if no more keys or error.
 */

extern p15_private_key_t* p15_copy_private_key(p15_private_key_t*);
/* Alocates new independent object and makes a smart copy. 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_private_key.
 */

extern void p15_free_private_key(p15_private_key_t**);
/* Frees an independent private key object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */



/* Public key functions */

extern int p15_add_public_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint8_t* key, uint16_t key_usage, int key_params,
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors);
/* Adds public key. 
 * About IDs - see p15_add_certificate
 * All input pointed structures are copied and should be freed manually if not needed.
 * key - 64 bytes of the key (RFC4491, 2.3.2)
 * key_usage - P15_KU_ ORed constants
 * key_params - P15_PKIPAR_ constant
 * start_date - (optional) - for information about the key lifetime
 * end_date - (optional)
 * Ret: 0-ok, 1-if only warnings, P15_RC_ const for main error.
 * See also p15_set_format notes above.
 */

/* See p15_set_pkey_subject above */

extern int p15_rm_public_key(p15_ctx_t* ctx, p15_id_t* id);
/* Removes private key by id.
 * Ret: P15_RC_ const
 */

extern int p15_find_public_keys_by_subject(p15_ctx_t* ctx, p15_der_t* subject, 
	p15_id_t** found);
/* Same as p15_find_certs_by_subject */

extern p15_public_key_t* p15_get_public_key(p15_ctx_t* ctx, p15_id_t* id);
/* Returns read-only structure of public key, found by id. 
 * Or NULL if not found or error. 
 * Returned structure is a part of ctx and should not be freed manually. 
 */

extern p15_public_key_t* p15_get_next_public_key(p15_ctx_t* ctx, p15_object_t** it);
/* Iterates next private key from the global list. 
 * *it=NULL - to start iteration.
 * Returns NULL if no more keys or error.
 */

extern p15_public_key_t* p15_copy_public_key(p15_public_key_t*);
/* Alocates new independent object and makes a smart copy. 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_public_key.
 */

extern void p15_free_public_key(p15_public_key_t**);
/* Frees an independent public key object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */


/* Secret key functions */


extern int p15_add_secret_key(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	p15_info_t** errors,
	int mask_count, uint8_t* key, ...);
/* Adds secret key. 
 * If id set to NULL, new unique ID is assigned. If 
 * id - if NULL - new id is created and returned to newly allocated *newid (if newid!=NULL).
 * If id is specified - given key stored with given id. 
 * ID for secret key should be unique (among all objects). 
 * (id struct is copied, and the source id struct should be freed if not needed).
 * All input pointed structures are copied and should be freed manually if not needed.
 * key_usage - P15_KU_ ORed constants
 * key_access_flags - may be P15_KA_ABSENT, P15_KA_PRESENT, other P15_KA_ ORed constants.
 * P15_KA_DEFAULT is recommended. See above about P15_KA_ constants.
 * key_params - P15_UZ89_ constant
 * start_date - (optional) - for information about the key lifetime
 * end_date - (optional)
 * mask_count - count of key+masks. 1-key only, 2-key w/mask, 3-key w/2 masks,...
 * key - 32 bytes of the (masked) key (little-endian)
 * Masking is XOR-based.
 * If mask_count > 1, uint8_t* pointers should follow key arg, pointing to masks.
 * Ret: 0-ok, 1-if only warnings, P15_RC_ const for main error.
 * See also p15_set_format notes above.
 */

extern int p15_add_secret_key2(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid,
	char* label, uint16_t key_usage, uint8_t key_access_flags, int key_params, 
	p15_utctime_t* start_date, p15_utctime_t* end_date,
	int mask_count, uint8_t** key_masks,
	p15_info_t** errors);
/* Same as p15_add_secret_key, except key_masks - is an array of key masks:
 * key_mask[0] - masked key
 * key_mask[1] - mask 1
 * key_mask[2] - mask 2 ...
 * All input pointed structures are copied and should be freed manually if not needed.
 */

extern int p15_remask_secret_key(p15_ctx_t* ctx, p15_secret_key_t* key, int mask_count);
/* Masks/umasks/remasks a secret key. 
 * Random generator should be initialized.
 * mask_count should be > 0.
 * If mask_count==1 - the key is unmasked.
 * Ret: P15_RC_ const.
 */ 


extern int p15_rm_secret_key(p15_ctx_t* ctx, p15_id_t* id);
/* Removes secret key by id.
 * Ret: P15_RC_ const
 */


extern p15_secret_key_t* p15_get_secret_key(p15_ctx_t* ctx, p15_id_t* id);
/* Returns read-only structure of secret key, found by id. 
 * Or NULL if not found or error. 
 * Returned structure is a part of ctx and should not be freed manually. 
 * If key is not decrypted (wrong password) - key_masks==NULL.
 */

extern p15_secret_key_t* p15_get_next_secret_key(p15_ctx_t* ctx, p15_object_t** it);
/* Iterates next secret key from the global list. 
 * *it=NULL - to start iteration.
 * Returns NULL if no more keys or error.
 */

extern p15_secret_key_t* p15_copy_secret_key(p15_secret_key_t*);
/* Alocates new independent object and makes a smart copy. 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_secret_key.
 */

extern void p15_free_secret_key(p15_secret_key_t**);
/* Frees an independent secret key object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */



/* CRL functions */

extern int p15_add_crl(p15_ctx_t* ctx, p15_id_t* id, p15_id_t** newid, 
	char* label, uint8_t* crl, int crllen, int mult_crls_per_issuer, 
	p15_info_t** errors);
/* Adds CRL.
 * All input pointed structures are copied and should be freed manually if not needed.
 * crl should be in DER.
 * If mult_crls_per_issuer==0 - only one CRL per one Issuer allowed.
 * If mult_crls_per_issuer==1 - multiple CRLs per one Issuer allowed, but they must
 * differ in thisUpdate/nextUpdate time stamps.
 * See above about other args.
 */


extern int p15_rm_crl(p15_ctx_t* ctx, p15_id_t* id);
/* Removes CRL by id
 * Ret: P15_RC_ const
 */


extern int p15_find_crl_by_issuer(p15_ctx_t* ctx, p15_der_t* issuer,
	p15_id_t** found);
/* Same as p15_find_certs_by_subject. */

extern p15_crl_t* p15_get_crl(p15_ctx_t* ctx, p15_id_t* id);
/* Returns read-only CRL structure, found by id. Or NULL if not found or error.
 * Returned structure is a part of ctx and should not be freed manually. 
 */

extern p15_crl_t* p15_get_next_crl(p15_ctx_t* ctx, p15_object_t** it);
/* Iterates next CRL from the global list. 
 * *it=NULL - to start iteration.
 * Returns NULL if no more CRLs or error.
 */

extern p15_crl_t* p15_copy_crl(p15_crl_t*);
/* Alocates new independent object and makes a smart copy. 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_crl.
 */

extern void p15_free_crl(p15_crl_t**);
/* Frees an independent CRL object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */


/* OID functions */

extern p15_oid_t* p15_new_oid_from_str(const char* oidstr);
/* Allocates new oid object created from a string like "1.23.456.7890"
 * Ret: NULL if error
 */

extern p15_oid_t* p15_new_oid_from_arcs(uint32_t* arcs, int num);
/* Allocates new oid object created from array of oid parts. num is the array length
 * Ret: NULL if error
 */

extern p15_oid_t* p15_copy_oid(p15_oid_t* oid);
/* Copies the oid object.
 * Ret: NULL if error
 */

extern void p15_free_oid(p15_oid_t** oid);
/* Frees the oid object */

extern int p15_oid_cmp(p15_oid_t* oid1, p15_oid_t* oid2);
/* Compares OIDs.
 * Ret: 0-equal, 1-not equal, P15_RC_INVALID_ARGS
 */

extern int p15_oid_to_str(p15_oid_t* oid, char* str, int maxlen);
/* Makes a string like "1.23.456.7890" from the oid.
 * Ret: 0-ok, P15_RC_INVALID_ARGS, P15_RC_BUFFER_TOO_SMALL
 */



/* Abstract data objects functions */

extern int p15_add_oiddo(p15_ctx_t* ctx, char* label, 
	const char* oidst, p15_oid_t* oid, 
	uint8_t* data, size_t len, int to_encrypt, p15_info_t** errors);
/* Adds abstract data object identified by oid.
 * One of 'oidst' or 'oid' parameters should be NULL.
 * All input data is copied.
 * if to_encrypt==1 - the data object is encrypted in the token
 * Ret: P15_RC_ const
 */

extern p15_oiddo_t* p15_get_next_oiddo(p15_ctx_t* ctx, p15_object_t** it, 
	p15_oid_t* oid);
/* Iterates to next oiddo object. 
 * *it=NULL - to start iteration.
 * If oid==NULL - iterates all met oiddo objects.
 * If oid!=NULL - iterates only objects which match given oid.
 * Returns NULL if no more objects found or error
 */

extern p15_oiddo_t* p15_copy_oiddo(p15_oiddo_t*);
/* Alocates new independent object and makes a smart copy. 
 * Returns NULL if error.
 * The returned object should be freed by p15_free_oiddo.
 */

extern void p15_free_oiddo(p15_oiddo_t**);
/* Frees an independent object.
 * This function MUST NOT be called on the objects from p15_ctx_t!
 */



/* Token output function */

extern int p15_save_token(p15_ctx_t* ctx, uint8_t* token_out, size_t* token_len, 
	uint8_t* password, int passlen, p15_info_t** errors);
/* Saves ready token to token_out buffer, constructed from the ctx data.
 * *token_len should contain the token_out buffer size.
 * Actual token size is returned back to token_len.
 * If token_out is smaller - P15_RC_BUFFER_TOO_SMALL error is returned and demanded
 * token size returned to token_len.
 * If *token_len==0 - P15_RC_BUFFER_TOO_SMALL is NOT returned, but P15_RC_OK.
 * Random generator should be initialized (see p15_init_random).
 * New content-encryption session keys are created by rng.
 * New random init value is written to the token. (See p15_random_setup).
 * Private keys are remasked if to_remask flag is set. (See p15_remask_setup).
 * If password==NULL - empty password is used for key-encryption-key generation
 * If errors!=NULL - error list is allocated (or appended) with detailed errors
 * (*errors must be freed by p15_free_infos after the call).
 * Returns: 0-ok, !0-P15_RC_ const of most significant error.
 * See also p15_set_format notes above.
 */


#ifdef __cplusplus
} /* extern "C" */
#endif 



#endif

