#include <stdint.h>
#include <string.h>
#include <strings.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include "pkcs15container.h"



/*****************************************************************************/
void Usage()
{
	printf("PKCS15 2.0->1.0 converter\n");
	printf("Usage: p15conv <infile> <outfile>\n");
}
/*****************************************************************************/
int main(int argc, char** argv)
{
	p15_ctx_t* ctx = NULL;
	int f = -1;
	int grc = 1;
	int rc;
	uint8_t* buf = NULL;
	struct stat sb;
	size_t sz = 0;

	if (argc != 3) {
		Usage();
		return 1;
	}

	ctx = p15_new_ctx();

	printf("Loading file %s\n", argv[1]);

	f = open(argv[1], O_RDONLY);
	if (f < 0) {
		printf("Failed to open file\n");
		goto end;
	}
	bzero(&sb, sizeof(sb));
	if (fstat(f, &sb)) {
		printf("Failed to stat file\n");
		goto end;
	}
	sz = sb.st_size;
	buf = malloc(sz);
	if (read(f, buf, sz) != sz) {
		printf("Failed to read file\n");
		goto end;
	}
	close(f);
	f = -1;

	rc = p15_load_token(ctx, buf, sz, NULL, 0, NULL);
	if (rc == P15_RC_WRONG_PASSWORD) {
		printf("Container is protected by password. Not impemented yet.\n");
		goto end;
	}
	if (rc) {
		printf("Error loading the container. Code %d.\n", rc);
		goto end;
	}
		
	rc = p15_set_format(ctx, P15_FORMAT_2001);
	if (rc) {
		printf("Error downgrading format. Code %d\n", rc);
		goto end;
	}
	free(buf);
	buf = NULL;

	sz = 0;
	rc = p15_save_token(ctx, NULL, &sz, NULL, 0, NULL);
	if (rc > 1) {
		printf("Error detecting output container size. Code %d\n", rc);
		goto end;
	}
	buf = malloc(sz);
	rc = p15_save_token(ctx, buf, &sz, NULL, 0, NULL);
	if (rc > 1) {
		printf("Error saving token. Code %d\n", rc);
		goto end;
	}

	printf("Saving token to %s file\n", argv[2]);
	unlink(argv[2]);
	f = open(argv[2], O_CREAT | O_WRONLY, 0644);
	if (f < 0) {
		printf("Failed to create file\n");
		goto end;
	}
	if (write(f, buf, sz) != sz) {
		printf("Failed to write file\n");
		goto end;
	}
	close(f);
	f = -1;

	grc = 0;
end:
	if (f > -1) close(f);
	p15_free_ctx(&ctx);
	free(buf);
	return grc;
}

